import warnings
from abc import abstractmethod
from typing import Iterable, Optional

import torch

from leaspy.exceptions import LeaspyModelInputError
from leaspy.io.data.dataset import Dataset
from leaspy.utils.typing import DictParamsTorch, KwargsType
from leaspy.utils.weighted_tensor import WeightedTensor
from leaspy.variables.dag import VariablesDAG
from leaspy.variables.specs import (
    Hyperparameter,
    IndividualLatentVariable,
    LatentVariableInitType,
    ModelParameter,
    NamedVariables,
    PopulationLatentVariable,
    VariableName,
    VariableNameToValueMapping,
)
from leaspy.variables.state import State, StateForkType

from .base import BaseModel

__all__ = ["StatefulModel"]


class StatefulModel(BaseModel):
    """Stateful models have an internal :class:`~leaspy.variables.State` to handle parameters and variables.

    Parameters
    ----------

    name : :obj:`str`
        The name of the model.

    Attributes
    ----------
    state : :class:`~leaspy.variables.State`
        The internal state of the model, which contains the variables and their values.
    tracked_variables : :obj:`set` [:obj:`str`]
        Set of variable names that are tracked by the model. These variables are not necessarily part of
        the model's state but are monitored for changes or updates. This can include variables that are
        relevant for the model's operation but not directly stored in the state.

    """

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self._state: Optional[State] = None
        self.tracked_variables: set[str] = set()

    def track_variable(self, variable: VariableName) -> None:
        """Track a variable by its name.

        Parameters
        -------
        variable : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to track. This variable will be monitored for changes or updates.
        """

        self.tracked_variables.add(variable)

    def track_variables(self, variables: Iterable[VariableName]) -> None:
        """
        Track multiple variables by their names.

        Parameters
        ----------
        variables : :obj:`Iterable` [:class:`~leaspy.variables.specs.VariableName`]
            An iterable containing the names of the variables to track. Each variable will be monitored
            for changes or updates.

        """
        for variable in variables:
            self.track_variable(variable)

    def untrack_variable(self, variable: VariableName) -> None:
        """Untrack a variable by its name.

        Parameters
        -------
        variable : :class:`~leaspy.variables.specs.VariableName`
            The name of the variable to untrack. This variable will no longer be monitored for changes or updates.
        """
        self.tracked_variables.remove(variable)

    def untrack_variables(self, variables: Iterable[VariableName]) -> None:
        """Untrack multiple variables by their names.

        Parameters
        ----------
        variables : :obj:`Iterable` [:class:`~leaspy.variables.specs.VariableName`]
            An iterable containing the names of the variables to untrack. Each variable will no longer be monitored
            for changes or updates.
        """
        for variable in variables:
            self.untrack_variable(variable)

    @property
    def state(self) -> State:
        """Get the internal state of the model.

        Returns
        -------
        State : :class:`~leaspy.variables.State`
            The internal state of the model, which contains the variables and their values.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If the model's state is not initialized yet.
        """
        if self._state is None:
            raise LeaspyModelInputError("Model state is not initialized yet")
        return self._state

    @state.setter
    def state(self, s: State) -> None:
        """Set the internal state of the model.
        This method allows to set the internal state of the model, which is an instance of

        Parameters
        ----------
        s : :class:`~leaspy.variables.State`
            The new state to set for the model.

        Raises
        ------
        LeaspyModelInputError
            If the provided state does not match the previous state in terms of DAG structure.
        """
        assert isinstance(s, State), "Provided state should be a valid State instance"
        if self._state is not None and s.dag is not self._state.dag:
            raise LeaspyModelInputError(
                "DAG of new state does not match with previous one"
            )
        # TODO? perform some clean-up steps for provided state (cf. `_terminate_algo` of MCMC algo)
        self._state = s

    @property
    def dag(self) -> VariablesDAG:
        """Get the underlying DAG of the model's state.

        Returns
        -------
        : :class:`~leaspy.variables.dag.VariablesDAG`
            The directed acyclic graph (DAG) representing the model's variables and their relationships

        """
        return self.state.dag

    @property
    def hyperparameters_names(self) -> tuple[VariableName, ...]:
        """Get the names of the model's hyperparameters.

        Returns
        -------
        : :obj:`tuple` [:class:`~leaspy.variables.specs.VariableName`, others...]
            A tuple containing the names of the model's hyperparameters.
        """
        return tuple(self.dag.sorted_variables_by_type[Hyperparameter])

    @property
    def parameters_names(self) -> tuple[VariableName, ...]:
        """Get the names of the model's parameters.

        Returns
        -------
        : :obj:`tuple` [:class:`~leaspy.variables.specs.VariableName`, others...]
            A tuple containing the names of the model's parameters.
        """
        return tuple(self.dag.sorted_variables_by_type[ModelParameter])

    @property
    def population_variables_names(self) -> tuple[VariableName, ...]:
        """Get the names of the population latent variables.

        Returns
        -------
        : :obj:`tuple` [:class:`~leaspy.variables.specs.VariableName`, ...]
            A tuple containing the names of the population latent variables.
        """
        return tuple(self.dag.sorted_variables_by_type[PopulationLatentVariable])

    @property
    def individual_variables_names(self) -> tuple[VariableName, ...]:
        """Get the names of the individual latent variables.

        Returns
        -------
        : :obj:`tuple` [:class:`~leaspy.variables.specs.VariableName`, ...]
            A tuple containing the names of the individual latent variables.
        """
        return tuple(self.dag.sorted_variables_by_type[IndividualLatentVariable])

    @property
    def parameters(self) -> DictParamsTorch:
        """Dictionary of values for model parameters.

        Returns
        -------
        : :class:`~leaspy.utils.typing.DictParamsTorch`
            A dictionary mapping parameter names to their values (as tensors).
        """
        return {p: self._state[p] for p in self.parameters_names}

    @property
    def hyperparameters(self) -> DictParamsTorch:
        """Dictionary of values for model hyperparameters.

        Returns
        -------
        : :class:`~leaspy.utils.typing.DictParamsTorch`
            A dictionary mapping hyperparameter names to their values (as tensors).
        """
        return {p: self._state[p] for p in self.hyperparameters_names}

    def initialize(self, dataset: Optional[Dataset] = None) -> None:
        """Overloads base model initialization (in particular to handle internal model State).

        <!> We do not put data variables in internal model state at this stage (done in algorithm)

        Parameters
        ----------
        dataset : :class:`~leaspy.io.data.dataset.Dataset`, optional
            Input dataset from which to initialize the model.
        """
        super().initialize(dataset=dataset)
        self._initialize_state()
        if not dataset:
            return
        # WIP: design of this may be better somehow?
        with self._state.auto_fork(None):
            self._initialize_model_parameters(dataset)
            self._state.put_population_latent_variables(
                LatentVariableInitType.PRIOR_MODE
            )

    def _initialize_state(self) -> None:
        """Initialize the internal state of model, as well as the underlying DAG.

        Note that all model hyperparameters (dimension, source_dimension, ...) should be defined
        in order to be able to do so.
        """
        if self._state is not None:
            raise LeaspyModelInputError("Trying to initialize the model's state again")
        self.state = State(
            VariablesDAG.from_dict(self.get_variables_specs()),
            auto_fork_type=StateForkType.REF,
        )
        self.state.track_variables(self.tracked_variables)

    def _initialize_model_parameters(self, dataset: Dataset) -> None:
        """Initialize model parameters (in-place, in `_state`).

        The method also checks that the model parameters whose initial values
        were computed from the dataset match the expected model parameters from
        the specifications (i.e. the nodes of the DAG of type 'ModelParameter').

        If there is a mismatch, the method raises a ValueError because there is
        an inconsistency between the definition of the model and the way it computes
        the initial values of its parameters from a dataset.

        Parameters
        ----------
        dataset : :class:`~leaspy.io.data.dataset.Dataset`
            The dataset to use to compute initial values for the model parameters.
        """
        model_parameters_initialization = (
            self._compute_initial_values_for_model_parameters(dataset)
        )
        model_parameters_spec = self.dag.sorted_variables_by_type[ModelParameter]
        if set(model_parameters_initialization.keys()) != set(model_parameters_spec):
            raise ValueError(
                "Model parameters created at initialization are different "
                "from the expected model parameters from the specs:\n"
                f"- From initialization: {sorted(list(model_parameters_initialization.keys()))}\n"
                f"- From Specs: {sorted(list(model_parameters_spec))}\n"
            )
        for (
            model_parameter_name,
            model_parameter_variable,
        ) in model_parameters_spec.items():
            model_parameter_initial_value = model_parameters_initialization[
                model_parameter_name
            ]
            if not isinstance(
                model_parameter_initial_value, (torch.Tensor, WeightedTensor)
            ):
                try:
                    model_parameter_initial_value = torch.tensor(
                        model_parameter_initial_value, dtype=torch.float
                    )
                except ValueError:
                    raise ValueError(
                        f"The initial value for model parameter '{model_parameter_name}' "
                        "should be a tensor, or a weighted tensor.\nInstead, "
                        f"{model_parameter_initial_value} of type {type(model_parameter_initial_value)} "
                        "was received and cannot be casted to a tensor.\nPlease verify this parameter "
                        "initialization code."
                    )
            self._state[model_parameter_name] = model_parameter_initial_value.expand(
                model_parameter_variable.shape
            )

    def load_parameters(self, parameters: KwargsType) -> None:
        """Instantiate or update the model's parameters.

        It assumes that all model hyperparameters are defined.

        Parameters
        ----------
        parameters : :class:`~leaspy.utils.typing.KwargsType`]
            Contains the model's parameters.
        """
        from .utilities import val_to_tensor

        if self._state is None:
            self._initialize_state()

        # TODO: a bit dirty due to hyperparams / params mix (cf. `.parameters` property note)

        params_names = self.parameters_names
        missing_params = set(params_names).difference(parameters)
        if len(missing_params):
            warnings.warn(f"Missing some model parameters: {missing_params}")
        extra_vars = set(parameters).difference(self.dag)
        if len(extra_vars):
            raise LeaspyModelInputError(f"Unknown model variables: {extra_vars}")
        # TODO: check no DataVariable provided???
        # extra_params = set(parameters).difference(cur_params)
        # if len(extra_params):
        #    # e.g. mixing matrix, which is a derived variable - checking their values only
        #    warnings.warn(f"Ignoring some provided values that are not model parameters: {extra_params}")
        # update parameters first (to be able to check values of derived variables afterwards)
        provided_params = {
            p: val_to_tensor(parameters[p], self.dag[p].shape)
            for p in params_names
            if p in parameters
        }
        for p, val in provided_params.items():
            # TODO: WeightedTensor? (e.g. batched `deltas`)
            self._state[p] = val

        # derive the population latent variables from model parameters
        # e.g. to check value of `mixing_matrix` we need `v0` and `betas` (not just `log_v0` and `betas_mean`)
        self._state.put_population_latent_variables(LatentVariableInitType.PRIOR_MODE)

        # check equality of other values (hyperparameters or linked variables)
        for parameter_name, parameter_value in parameters.items():
            if parameter_name in provided_params:
                continue
            # TODO: a bit dirty due to hyperparams / params mix (cf. `.parameters` property note)
            try:
                current_value = self._state[parameter_name]
            except Exception as e:
                raise LeaspyModelInputError(
                    f"Impossible to compare value of provided value for {parameter_name} "
                    "- not computable given current state"
                ) from e
            parameter_value = val_to_tensor(
                parameter_value, getattr(self.dag[parameter_name], "shape", None)
            )
            assert (
                parameter_value.shape == current_value.shape,
                (parameter_name, parameter_value.shape, current_value.shape),
            )
            # TODO: WeightedTensor? (e.g. batched `deltas``)
            assert (
                torch.allclose(parameter_value, current_value, atol=1e-4),
                (parameter_name, parameter_value, current_value),
            )

    @abstractmethod
    def get_variables_specs(self) -> NamedVariables:
        """Return the specifications of the variables (latent variables,
        derived variables, model 'parameters') that are part of the model.

        Returns
        -------
        NamedVariables : :class:`~leaspy.variables.specs.NamedVariables`
            The specifications of the model's variables.
        """
        raise NotImplementedError()

    @abstractmethod
    def _compute_initial_values_for_model_parameters(
        self, dataset: Dataset
    ) -> VariableNameToValueMapping:
        """Compute initial values for model parameters.

        Parameters
        ----------
        dataset : :class:`~leaspy.io.data.dataset.Dataset`
            The dataset to use to compute initial values for the model parameters.

        Returns
        -------
        : :class:`~leaspy.utils.typing.Any`
            A dictionary mapping variable names to their initial values.
        """
        raise NotImplementedError()

    def move_to_device(self, device: torch.device) -> None:
        """Move a model and its relevant attributes to the specified :class:`torch.device`.

        Parameters
        ----------
        device : :obj:`torch.device`
            The device to which the model and its attributes should be moved.
        """
        if self._state is None:
            return
        self._state.to_device(device)
        for hp in self.hyperparameters_names:
            self._state.dag[hp].to_device(device)



import torch

__all__ = [
    "compute_individual_parameter_std_from_sufficient_statistics",
]


def compute_individual_parameter_std_from_sufficient_statistics(
    state: dict[str, torch.Tensor],
    individual_parameter_values: torch.Tensor,
    individual_parameter_sqr_values: torch.Tensor,
    *,
    individual_parameter_name: str,
    dim: int,
    **kws,
):
    """
    Maximization rule, from the sufficient statistics, of the standard-deviation of Gaussian prior for individual latent variables.

    Parameters
    ----------
    state : :obj:`dict`[:obj:`str`, :class:`torch.Tensor`]
        The current state object that holds all the variables
    individual_parameter_values : :class:`torch.Tensor`
        Tensor containing individual parameter values, used to compute current means.
    individual_parameter_sqr_values : :class:`torch.Tensor`
        Tensor containing squared individual parameter values, used to compute variances.
    individual_parameter_name : :obj:`str`
        The name of the individual parameter for which to compute the std.
    dim : :obj:`int`
        The dimension along which to compute the mean and variance
        
    Returns
    -------
    :class:`torch.Tensor`
        The updated standard deviation of the Gaussian prior for the individual parameter
    """
    from leaspy.models.utilities import compute_std_from_variance

    individual_parameter_old_mean = state[f"{individual_parameter_name}_mean"]
    individual_parameter_current_mean = torch.mean(individual_parameter_values, dim=dim)
    individual_parameter_variance_update = (
        torch.mean(individual_parameter_sqr_values, dim=dim)
        - 2 * individual_parameter_old_mean * individual_parameter_current_mean
    )
    individual_parameter_variance = (
        individual_parameter_variance_update + individual_parameter_old_mean**2
    )
    return compute_std_from_variance(
        individual_parameter_variance, varname=f"{individual_parameter_name}_std", **kws
    )



from importlib import import_module

__version__ = "2.0.0-dev"
dtype = "float32"

pkg_deps = [
    "torch",
    "numpy",
    "pandas",
    "scipy",  # core
    "sklearn",
    "joblib",  # parallelization / ML utils
    "statsmodels",  # LME benchmark only
    "matplotlib",  # plots
]

__watermark__ = {
    "leaspy": __version__,
    **{pkg_name: import_module(pkg_name).__version__ for pkg_name in pkg_deps},
}

del pkg_deps



from .base import (
    AlgorithmName,
    AlgorithmType,
    BaseAlgorithm,
    algorithm_factory,
    get_algorithm_class,
    get_algorithm_type,
)
from .settings import AlgorithmSettings, OutputsSettings, algo_default_data_dir

__all__ = [
    "BaseAlgorithm",
    "AlgorithmSettings",
    "OutputsSettings",
    "AlgorithmType",
    "AlgorithmName",
    "get_algorithm_type",
    "algorithm_factory",
    "get_algorithm_class",
    "algo_default_data_dir",
]



from ._functions import (
    Exp,
    Identity,
    MatMul,
    Mean,
    OrthoBasis,
    Prod,
    Sqr,
    Std,
    Sum,
    SumDim,
)
from ._named_input_function import NamedInputFunction
from ._utils import get_named_parameters

__all__ = [
    "Exp",
    "get_named_parameters",
    "Identity",
    "MatMul",
    "Mean",
    "NamedInputFunction",
    "OrthoBasis",
    "Prod",
    "Sqr",
    "Std",
    "Sum",
    "SumDim",
]



import torch

from leaspy.exceptions import LeaspyModelInputError

__all__ = [
    "compute_orthonormal_basis",
]


def compute_orthonormal_basis(
    dgamma_t0: torch.Tensor, G_metric: torch.Tensor, *, strip_col: int = 0
):
    """
    Householder decomposition, adapted for a non-Euclidean inner product defined by:
    (1) :math:`< x, y >Metric(p) = < x, G(p) y >Eucl = xT G(p) y`, where:
    :math:`G(p)` is the symmetric positive-definite (SPD) matrix defining the metric at point `p`.

    The Euclidean case is the special case where `G` is the identity matrix.
    Product-metric is a special case where `G(p)` is a diagonal matrix (identified to a vector)
    whose components are all > 0.

    It is used to compute and set in-place the ``orthonormal_basis`` attribute
    given the time-derivative of the geodesic at initial time and the `G_metric`.
    The first component of the full orthonormal basis is a vector collinear `G_metric x dgamma_t0` that we get rid of.

    The orthonormal basis we construct is always orthonormal for the Euclidean canonical inner product.
    But all (but first) vectors of it lie in the sub-space orthogonal (for canonical inner product) to `G_metric * dgamma_t0`
    which is the same thing that being orthogonal to `dgamma_t0` for the inner product implied by the metric.

    [We could do otherwise if we'd like a full orthonormal basis, w.r.t. the non-Euclidean inner product.
    But it'd imply to compute G^(-1/2) & G^(1/2) which may be computationally costly in case we don't have direct access to them
    (for the special case of product-metric it is easy - just the component-wise inverse (sqrt'ed) of diagonal)
    TODO are there any advantages/drawbacks of one method over the other except this one?
    TODO are there any biases between features when only considering Euclidean orthonormal basis?]

    Parameters
    ----------
    dgamma_t0 : :class:`torch.FloatTensor` 1D
        Time-derivative of the geodesic at initial time.
        It may also be a vector collinear to it without any change to the result.

    G_metric : scalar, `torch.FloatTensor` 0D, 1D or 2D-square
        The `G(p)` defining the metric as referred in equation (1) just before :
            * If 0D (scalar): `G` is proportional to the identity matrix
            * If 1D (vector): `G` is a diagonal matrix (diagonal components > 0)
            * If 2D (square matrix): `G` is general (SPD)

    strip_col : :obj:`int` in 0..model_dimension-1 (default 0)
        Which column of the basis should be the one collinear to `dgamma_t0` (that we get rid of)

    Returns
    -------
    torch.Tensor of shape (dimension, dimension - 1)

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        if incoherent metric `G_metric`
    """

    assert (
        dgamma_t0.ndim == 1
    ), f"Expecting vector for velocities but got {dgamma_t0.ndim}D tensor"
    (dimension,) = dgamma_t0.shape

    ## enforce `G_metric` to be a tensor
    # if not isinstance(G_metric, torch.Tensor):
    #    G_metric = torch.tensor(G_metric) # convert from scalar...

    # compute the vector that others columns should be orthogonal to, w.r.t canonical inner product
    G_shape = G_metric.shape
    if len(G_shape) == 0:  # 0D
        if G_metric.item() <= 0:
            raise LeaspyModelInputError("Incoherent negative scalar metric.")
        dgamma_t0 = G_metric.item() * dgamma_t0  # homothetic
    elif len(G_shape) == 1:  # 1D
        if not (G_metric > 0).all():
            raise LeaspyModelInputError("Incoherent 1D metric with negative values.")
        if G_shape != (dimension,):
            raise LeaspyModelInputError(
                f"Incoherent 1D metric size: {G_shape[0]} != {dimension}."
            )
        dgamma_t0 = G_metric * dgamma_t0  # component-wise multiplication of vectors
    elif len(G_shape) == 2:  # matrix (general case)
        # no check on positivity of matrix to remain light
        if G_shape != (dimension, dimension):
            raise LeaspyModelInputError(
                f"Incoherent 2D metric shape: {G_shape} != {(dimension, dimension)}."
            )
        dgamma_t0 = G_metric @ dgamma_t0  # matrix multiplication
    else:
        raise LeaspyModelInputError(
            "Unexpected metric of dim > 2 when computing orthonormal basis."
        )

    """
    Automatically choose the best column to strip?
    <!> Not a good idea because it could fluctuate over iterations making mixing_matrix unstable!
        (betas should slowly readapt to the permutation...)
    #strip_col = dgamma_t0.abs().argmax().item()
    #strip_col = v_metric_normalization.argmin().item()
    """

    assert isinstance(strip_col, int) and 0 <= strip_col < dimension, (
        strip_col,
        dimension,
    )
    ej = torch.zeros_like(dgamma_t0)
    ej[strip_col] = 1.0

    alpha = -torch.sign(dgamma_t0[strip_col]) * torch.norm(dgamma_t0)
    u_vector = dgamma_t0 - alpha * ej
    v_vector = u_vector / torch.norm(u_vector)

    ## Classical Householder method (to get an orthonormal basis for the canonical inner product)
    ## Q = I_n - 2 v • v'
    q_matrix = torch.eye(dimension) - 2 * v_vector.view(-1, 1) * v_vector

    # concat columns (get rid of the one collinear to `dgamma_t0`)
    return torch.cat((q_matrix[:, :strip_col], q_matrix[:, strip_col + 1 :]), dim=1)



import torch

from leaspy.io.data.dataset import Dataset
from leaspy.utils.weighted_tensor import WeightedTensor
from leaspy.variables.distributions import Bernoulli
from leaspy.variables.specs import VariableInterface

from ._base import ObservationModel

__all__ = ["BernoulliObservationModel"]


class BernoulliObservationModel(ObservationModel):
    """
    Observation model for binary outcomes using a Bernoulli distribution.

    This model expects binary-valued observations and uses a Bernoulli distribution
    to define the likelihood. It assumes the response variable is named `"y"`.

    Parameters
    ----------
    **extra_vars : VariableInterface
        Optional extra variables required by the model. These are passed to the
        parent `ObservationModel` class and can be used for conditioning the likelihood.

    Attributes
    ----------
    string_for_json : :obj:`str`
        A static string identifier used for serialization.
    """
    string_for_json = "bernoulli"

    def __init__(
        self,
        **extra_vars: VariableInterface,
    ):
        super().__init__(
            name="y",
            getter=self.y_getter,
            dist=Bernoulli("model"),
            extra_vars=extra_vars,
        )

    @staticmethod
    def y_getter(dataset: Dataset) -> WeightedTensor:
        """
        Extracts and validates the observation values and associated mask from a dataset.

        Parameters
        ----------
        dataset : :class:`.Dataset`
            A dataset object containing `values` and `mask` attributes.

        Returns
        -------
        :class:`.WeightedTensor`
            A tensor containing the observed binary values along with a boolean mask
            indicating which entries are valid.

        Raises
        ------
        ValueError
            If either `dataset.values` or `dataset.mask` is `None`, indicating that
            the dataset is improperly initialized.
        """
        if dataset.values is None or dataset.mask is None:
            raise ValueError(
                "Provided dataset is not valid. "
                "Both values and mask should be not None."
            )
        return WeightedTensor(dataset.values, weight=dataset.mask.to(torch.bool))



from __future__ import annotations

import operator
from dataclasses import dataclass
from typing import Callable, Generic, Optional, Tuple, TypeVar, Union

import torch

__all__ = [
    "WeightedTensor",
    "TensorOrWeightedTensor",
]


VT = TypeVar("VT")


@dataclass(frozen=True)
class WeightedTensor(Generic[VT]):
    """
    A torch.tensor, with optional (non-negative) weights (0 <-> masked).

    Parameters
    ----------
    value : :obj:`torch.Tensor`
        Raw values, without any mask.
    weight : :obj:`torch.Tensor`, optional
        Relative weights for values.
        Default: None

    Attributes
    ----------
    value : :obj:`torch.Tensor`
        Raw values, without any mask.
    weight : :obj:`torch.Tensor`
        Relative weights for values.
        If weight is a tensor[bool], it can be seen as a mask (valid value <-> weight is True).
        More generally, meaningless values <-> indices where weights equal 0.
        Default: None
    """

    value: torch.Tensor
    weight: Optional[torch.Tensor] = None

    def __post_init__(self):
        """
        Post-initialization method to ensure that the value and weight tensors are properly initialized.

        Raises:
        ------
        AssertionError:
            - If `value` is a `WeightedTensor` (disallowed for initialization).
            - If `weight` is a `WeightedTensor` (disallowed for weights).
            - If `weight` contains negative values.
            - If `weight` and `value` have mismatched shapes (no implicit broadcasting allowed).
            - If `weight` and `value` are on different devices.
        """
        if not isinstance(self.value, torch.Tensor):
            assert not isinstance(
                self.value, WeightedTensor
            ), "You should NOT init a `WeightedTensor` with another"
            object.__setattr__(self, "value", torch.tensor(self.value))
        if self.weight is not None:
            if not isinstance(self.weight, torch.Tensor):
                assert not isinstance(
                    self.weight, WeightedTensor
                ), "You should NOT use a `WeightedTensor` for weights"
                object.__setattr__(self, "weight", torch.tensor(self.weight))
            assert (self.weight >= 0).all(), "Weights must be non-negative"
            # we forbid implicit broadcasting of weights for safety
            assert (
                self.weight.shape == self.value.shape
            ), f"Bad shapes: {self.weight.shape} != {self.value.shape}"
            assert (
                self.weight.device == self.value.device
            ), f"Bad devices: {self.weight.device} != {self.value.device}"

    @property
    def weighted_value(self) -> torch.Tensor:
        """
        Get the weighted value tensor.
        This is the value tensor multiplied by the weight tensor.

        Returns
        -------
        :obj:`torch.Tensor`:
            The weighted value tensor.
            If weight is None, the value tensor is returned.
        """
        if self.weight is None:
            return self.value
        return self.weight * self.filled(0)

    def __getitem__(self, indices) -> WeightedTensor:
        """
        Get the weighted tensor at the specified indices.

        Parameters
        ----------
        indices : :obj:`torch.Tensor`
            The indices to get the weighted tensor at.

        Returns
        -------
        :class:`WeightedTensor`:
            A new `WeightedTensor` with the values and weights at the specified indices.
        """

        if self.weight is None:
            return WeightedTensor(self.value.__getitem__(indices), None)
        return WeightedTensor(
            self.value.__getitem__(indices), self.weight.__getitem__(indices)
        )

    def filled(self, fill_value: Optional[VT] = None) -> torch.Tensor:
        """Return the values tensor with masked zeros filled with the specified value.

        Return the values tensor filled with `fill_value` where the `weight` is exactly zero.

        Parameters
        ----------
        fill_value : :obj:`VT`, optional
            The value to fill the tensor with for aggregates where weights were all zero.
            Default: None

        Returns
        -------
        :obj:`torch.Tensor`:
            The filled tensor.
            If `weight` or fill_value is None, the original tensor is returned.

        """
        if fill_value is None or self.weight is None:
            return self.value
        return self.value.masked_fill(self.weight == 0, fill_value)

    def valued(self, value: torch.Tensor) -> WeightedTensor:
        """
        Return a new WeightedTensor with same weight as self but with new value provided.

        Parameters
        ----------
        value : :obj:`torch.Tensor`
            The new value to be set.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new WeightedTensor with the same weight as self but with the new value provided.
        """
        return type(self)(value, self.weight)

    def map(
        self,
        func: Callable[[torch.Tensor], torch.Tensor],
        *args,
        fill_value: Optional[VT] = None,
        **kws,
    ) -> WeightedTensor:
        """Apply a function to the values tensor while preserving weights.

        The function is applied only to the values tensor after optionally filling
        zero-weight positions. The weights remain unchanged in the returned tensor.

        Parameters
        ----------
        func : Callable[[ :obj:`torch.Tensor` ], :obj:`torch.Tensor` ]
            The function to be applied to the values.
        *args :
            Positional arguments to be passed to the function.
        fill_value : :obj:`VT`, optional
            The value to fill the tensor with for aggregates where weights were all zero.
            Default: None
        **kws :
            Keyword arguments to be passed to the function.

        Returns
        -------
        :class:`WeightedTensor`:
            A new `WeightedTensor` with the result of the operation and the same weights.

        """
        return self.valued(func(self.filled(fill_value), *args, **kws))

    def map_both(
        self,
        func: Callable[[torch.Tensor], torch.Tensor],
        *args,
        fill_value: Optional[VT] = None,
        **kws,
    ) -> WeightedTensor:
        """Apply a function to both values and weights tensors.

        The same function is applied to both components of the weighted tensor.
        Zero-weight positions in the values tensor are filled before applying the function.

        Parameters
        ----------
        func : Callable[[ :obj`torch.Tensor` ], :obj:`torch.Tensor` ]
            The function to be applied to both values and weights.
        *args :
            Positional arguments to be passed to the function.
        fill_value : :obj:`VT`, optional
            The value to fill the tensor with for aggregates where weights were all zero.
            Default: None
        **kws :
            Keyword arguments to be passed to the function.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the result of the operation and the appropriate weights.
        """
        return type(self)(
            func(self.filled(fill_value), *args, **kws),
            func(self.weight, *args, **kws) if self.weight is not None else None,
        )

    def index_put(
        self,
        indices: Tuple[torch.Tensor, ...],  # of ints
        values: torch.Tensor,  # of VT
        *,
        accumulate: bool = False,
    ) -> WeightedTensor[VT]:
        """
        Out-of-place :func:`torch.index_put` on values (no modification of weights).

        Parameters
        ----------
        indices : :obj:`tuple` [ :obj:`torch.Tensor`, ...]
            The indices to put the values at.
        values : :obj:`torch.Tensor`
            The values to put at the specified indices.
        accumulate : :obj:`bool`, optional
            Whether to accumulate the values at the specified indices.
            Default: False

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor.WeightedTensor` [ :obj:`VT` ]
            A new :class:`~leaspy.utils.weighted_tensor.WeightedTensor` with the updated values and the same weights.
        """
        return self.map(
            torch.index_put, indices=indices, values=values, accumulate=accumulate
        )

    def wsum(self, *, fill_value: VT = 0, **kws) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Get the weighted sum of tensor together with sum of weights.

        <!> The result is NOT a `WeightedTensor` any more since weights are already taken into account.
        <!> We always fill values with 0 prior to weighting to prevent 0 * nan = nan that would propagate nans in sums.

        Parameters
        ----------
        fill_value : :obj:`VT`, optional
            The value to fill the sum with for aggregates where weights were all zero.
            Default: 0
        **kws
            Optional keyword-arguments for torch.sum (such as `dim=...` or `keepdim=...`)

        Returns
        -------

        :obj:`tuple` [ :obj:`torch.Tensor`, :obj:`torch.Tensor` ]:
        Tuple containing:
            - weighted_sum : :obj:`torch.Tensor`
                Weighted sum, with totally un-weighted aggregates filled with `fill_value`.
            - sum_weights : :obj:`torch.Tensor` (may be of other type than `cls.weight_dtype`)
                The sum of weights (useful if some average are needed).
        """
        weight = self.weight
        if weight is None:
            weight = torch.ones_like(self.value, dtype=torch.bool)
        weighted_values = weight * self.filled(0)
        weighted_sum = weighted_values.sum(**kws)
        sum_weights = weight.sum(**kws)
        return weighted_sum.masked_fill(sum_weights == 0, fill_value), sum_weights

    def sum(self, *, fill_value: VT = 0, **kws) -> torch.Tensor:
        """Compute weighted sum of values.

        For unweighted tensors, this is equivalent to regular :func:`torch.sum`.
        For weighted tensors, returns the same as the first element of wsum().

        Parameters
        ----------
        fill_value : :obj:`VT`, optional
            The value to fill the sum with for aggregates where weights were all zero.
            Default: 0
        **kws
            Optional keyword-arguments

        Returns
        -------
        :obj:`torch.Tensor`:
            The weighted sum, with totally un-weighted aggregates filled with `fill_value`.
        """
        if self.weight is None:
            # more efficient in this case
            return self.value.sum(**kws)
        return self.wsum(fill_value=fill_value, **kws)[0]

    def view(self, *shape) -> WeightedTensor[VT]:
        """Return a view of the weighted tensor with a different shape.

        Parameters
        ----------
        shape : :obj:`tuple` [ :obj:`int`, ...]
            The new shape to be set.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor.WeightedTensor` [ :obj:`VT` ]:
            A new :class:`~leaspy.utils.weighted_tensor.WeightedTensor` with the same weights but with the new shape provided.
        """
        return self.map_both(torch.Tensor.view, *shape)

    def expand(self, *shape) -> WeightedTensor[VT]:
        """Expand the weighted tensor to a new shape.

        Parameters
        ----------
        shape : :obj:`tuple` [ :obj:`int`, ...]
            The new shape to be set.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor.WeightedTensor` [ :obj:`VT` ]:
            A new :class:`~leaspy.utils.weighted_tensor.WeightedTensor` with the same weights but with the new shape provided.
        """
        return self.map_both(torch.Tensor.expand, *shape)

    def to(self, *, device: torch.device) -> WeightedTensor[VT]:
        """Move the weighted tensor to a different device.

        Parameters
        ----------
        device : :obj:`torch.device`
            The device to be set.

        Returns
        -------
        :obj:`WeightedTensor`[:obj:`VT]:
            A new `WeightedTensor` with the same weights but with the new device provided.
        """
        return self.map_both(torch.Tensor.to, device=device)

    def cpu(self) -> WeightedTensor[VT]:
        """Move the weighted tensor to CPU memory.

        Applies the `torch.Tensor.cpu()` operation to both the value tensor and
        weight tensor (if present), returning a new weighted tensor with all
        components on the CPU.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor.WeightedTensor` [ :obj:`VT` ]:
            A new :class:`~leaspy.utils.weighted_tensor.WeightedTensor` with the same weights but with the new device provided.
        """
        return self.map_both(torch.Tensor.cpu)

    def __pow__(self, exponent: Union[int, float]) -> WeightedTensor[VT]:
        """
        Apply the power of the tensor to the specified exponent.

        Parameters
        ----------
        exponent : :obj:`int` or :obj:`float`
            The exponent to be applied.

        Returns
        -------
        :obj:`WeightedTensor`[:obj:`VT]:
            A new `WeightedTensor` with the same weights but with the new exponent applied.
        """
        return self.valued(self.value**exponent)

    @property
    def shape(self) -> torch.Size:
        """Shape of the values tensor.

        Returns
        -------
        :obj:`torch.Size`:
            The shape of the values tensor.
        """
        return self.value.shape

    @property
    def ndim(self) -> int:
        """Number of dimensions of the values tensor.

        Returns
        -------
        :obj:`int`:
            The number of dimensions of the values tensor.
        """
        return self.value.ndim

    @property
    def dtype(self) -> torch.dtype:
        """Type of values.

        Returns
        -------
        :obj:`torch.dtype`:
            The type of values.
        """

        return self.value.dtype

    @property
    def device(self) -> torch.device:
        """Device of values.

        Returns
        -------
        :obj:`torch.device`:
            The device of values.
        """
        return self.value.device

    @property
    def requires_grad(self) -> bool:
        """Whether the values tensor requires gradients.

        Returns
        -------
        :obj:`bool`:
            Whether the values tensor requires gradients.
        """
        return self.value.requires_grad

    def abs(self) -> WeightedTensor:
        """Compute the absolute value of the weighted tensor.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor.WeightedTensor`
            A new `WeightedTensor` with the absolute value of the values tensor.
        """
        return self.__abs__()

    def all(self) -> bool:
        """Check if all values are non-zero.

        Returns
        -------
        :obj:`bool`:
            Whether all values are non-zero.
        """
        return self.value.all()

    def __neg__(self) -> WeightedTensor:
        """Compute the negative of the weighted tensor.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the negative of the values tensor.
        """
        return WeightedTensor(-1 * self.value, self.weight)

    def __abs__(self) -> WeightedTensor:
        """Compute the absolute value of the weighted tensor.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the absolute value of the values tensor, the weight stay the same.
        """
        return WeightedTensor(abs(self.value), self.weight)

    def __add__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the sum of the weighted tensor and another tensor.

        Returns a new weighted tensor containing:
        - The sum of the value tensors
        - The weight according to the following rules:
            - If both tensors have weights: weights must be identical
            - If only one tensor has weights: those weights are retained
            - If neither tensor has weights: result has no weights

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to be added to the weighted tensor.
        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the summed values

        """
        return _apply_operation(self, other, "add")

    def __radd__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the sum of another tensor and the weighted tensor.

        Equivalent to `__add__` but with operands reversed. See `__add__` for details.

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to be added to the weighted tensor.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the sum of the other tensor and the values tensor.
        """
        return _apply_operation(self, other, "add", reverse=True)

    def __sub__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the difference between the weighted tensor and another tensor.

        Returns a new weighted tensor containing:
            - The difference between this tensor's values and the other tensor's values
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to be subtracted from the weighted tensor.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the differences and appropriate weights.
        """
        return _apply_operation(self, other, "sub")

    def __rsub__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the difference between another tensor and the weighted tensor.

        Equivalent to `__sub__` but with operands reversed. See `__sub__` for details.

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to be subtracted from the weighted tensor.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` containing the differences and appropriate weights.
        """
        return _apply_operation(self, other, "sub", reverse=True)

    def __mul__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the product of the weighted tensor and another tensor.

        Returns a new weighted tensor containing:
            - The product of the value tensors
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to be multiplied with the weighted tensor.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` containing the products and appropriate weights
        """
        return _apply_operation(self, other, "mul")

    def __rmul__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the product of another tensor and the weighted tensor.

        Equivalent to `__mul__` but with operands reversed. See `__mul__` for details.

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to be multiplied with the weighted tensor.
        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` containing the products and appropriate weights
        """
        return _apply_operation(self, other, "mul", reverse=True)

    def __truediv__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the division of the weighted tensor by another tensor.

        Returns a new weighted tensor containing:
            - The quotient of the value tensors
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to divide the weighted tensor by.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` containing the quotients and appropriate weights.

        """
        return _apply_operation(self, other, "truediv")

    def __rtruediv__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the division of another tensor by the weighted tensor.

        Equivalent to `__truediv__` but with operands reversed. See `__truediv__` for details.

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to be divided by the weighted tensor.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` containing the quotients and appropriate weights.
        """
        return _apply_operation(self, other, "truediv", reverse=True)

    def __lt__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the less-than comparison between the weighted tensor and another tensor.

        Returns a new weighted tensor containing boolean values indicating where:
            - This tensor's values are less than the other tensor's values
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to compare against

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the result of the less-than comparison and appropriate weights.
        """
        return _apply_operation(self, other, "lt")

    def __le__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the less-than-or-equal-to comparison between the weighted tensor and another tensor.

        Returns a new weighted tensor containing boolean values indicating where:
            - This tensor's values are less than or equal to the other tensor's values
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to compare against.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the result of the less-than-or-equal-to comparison and appropriate weights.
        """
        return _apply_operation(self, other, "le")

    def __eq__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the equality comparison between the weighted tensor and another tensor.

        Returns a new weighted tensor containing boolean values indicating where:
            - This tensor's values equal the other tensor's values
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to compare against.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the result of the equality comparison and appropriate weights..
        """
        return _apply_operation(self, other, "eq")

    def __ne__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the not-equal-to comparison between the weighted tensor and another tensor.

        Returns a new weighted tensor containing boolean values indicating where:
            - This tensor's values differ from the other tensor's values
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to compare against.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the result of the not-equal-to comparison and appropriate weights..
        """
        return _apply_operation(self, other, "ne")

    def __gt__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the greater-than comparison between the weighted tensor and another tensor.
        Returns a new weighted tensor containing boolean values indicating where:
            - This tensor's values exceed the other tensor's values
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to compare against.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the result of the greater-than comparison and appropriate weights.
        """
        return _apply_operation(self, other, "gt")

    def __ge__(self, other: TensorOrWeightedTensor) -> WeightedTensor:
        """Compute the greater-than-or-equal-to comparison between the weighted tensor and another tensor.

        Returns a new weighted tensor containing boolean values indicating where:
            - This tensor's values are greater than or equal to the other tensor's values
            - The weight according to the same rules as __add__

        Parameters
        ----------
        other : class:`TensorOrWeightedTensor`
            The tensor to compare against.

        Returns
        -------
        :obj:`WeightedTensor`:
            A new `WeightedTensor` with the result of the greater-than-or-equal-to compariso and appropriate weights.
        """
        return _apply_operation(self, other, "ge")

    @staticmethod
    def get_filled_value_and_weight(
        t: TensorOrWeightedTensor[VT], *, fill_value: Optional[VT] = None
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Method to get tuple (value, weight) for both regular and weighted tensors.

        Parameters
        ----------
        t : class:`TensorOrWeightedTensor`
            The tensor to be converted.
        fill_value : :obj:`VT`, optional
            The value to fill the tensor with for aggregates where weights were all zero.
            Default: None

        Returns
        -------
        :obj:`Tuple`[:obj:`torch.Tensor`, Optional[:obj:`torch.Tensor`]]:
            Tuple containing:
            - value : :obj:`torch.Tensor`
                The filled tensor.
                If `weight` is None, the original tensor is returned.
            - weight : :obj:`torch.Tensor`, optional
                The weight tensor.
        """
        if isinstance(t, WeightedTensor):
            return t.filled(fill_value), t.weight
        else:
            if not isinstance(t, torch.Tensor):
                t = torch.tensor(t)
            return t, None


TensorOrWeightedTensor = Union[torch.Tensor, WeightedTensor[VT]]


def _apply_operation(
    a: WeightedTensor,
    b: TensorOrWeightedTensor,
    operator_name: str,
    reverse: bool = False,
) -> WeightedTensor:
    """
    Apply a binary operation on two tensors, with the first one being a `WeightedTensor`.
    The second one can be a `WeightedTensor` or a regular tensor.
    The operation is applied to the values of the tensors, and the weights are handled accordingly.

    Parameters
    ----------
    a : :class:`WeightedTensor`
        The first tensor, which is a `WeightedTensor`.
    b : class:`TensorOrWeightedTensor`
        The second tensor, which can be a `WeightedTensor` or a regular tensor.
    operator_name : :obj:`str`
        The name of the binary operation to be applied.
    reverse : :obj:`bool`, optional
        If True, the operation is applied in reverse order (b operator a).
        Default: False
    Returns
    -------
    :class:`WeightedTensor`:
        A new `WeightedTensor` with the result of the operation and the appropriate weights.

    Raises
    ------
    :exc:`NotImplementedError`
        If the operation is not implemented for the given combination of tensors.
    """

    operation = getattr(operator, operator_name)
    if isinstance(b, WeightedTensor):
        result_value = (
            operation(b.value, a.value) if reverse else operation(a.value, b.value)
        )
        if a.weight is None:
            if b.weight is None:
                return WeightedTensor(result_value)
            else:
                return WeightedTensor(
                    result_value,
                    b.weight.expand(result_value.shape).clone()
                    if b.weight.shape != result_value.shape
                    else b.weight.clone(),
                )
        else:
            if b.weight is None:
                return WeightedTensor(
                    result_value,
                    a.weight.expand(result_value.shape).clone()
                    if a.weight.shape != result_value.shape
                    else a.weight.clone(),
                )
            else:
                if not torch.equal(a.weight, b.weight):
                    raise NotImplementedError(
                        f"Binary operation '{operator_name}' on two weighted tensors is "
                        "not implemented when their weights differ."
                    )
                return WeightedTensor(
                    result_value,
                    a.weight.expand(result_value.shape).clone()
                    if a.weight.shape != result_value.shape
                    else a.weight.clone(),
                )
    result_value = operation(b, a.value) if reverse else operation(a.value, b)
    result_weight = None
    if a.weight is not None:
        return WeightedTensor(
            result_value,
            (
                a.weight.expand(result_value.shape).clone()
                if a.weight.shape != result_value.shape
                else a.weight.clone()
            ),
        )

    return WeightedTensor(result_value)



import numpy as np

from leaspy.utils.typing import KwargsType

from .base import BaseModel

__all__ = ["StatelessModel"]


class StatelessModel(BaseModel):
    """Stateless model do not use an internal state to keep track of variables.

    Parameters are stored in an internal dictionary.

    Parameters
    ----------
    name : :obj:`str`
        The name of the model.
    """

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self._parameters: KwargsType = {}

    @property
    def parameters(self):
        return self._parameters

    def load_parameters(self, parameters: KwargsType) -> None:
        """Instantiate or update the model's parameters.

        Parameters
        ----------
        parameters : :obj:`dict`
            Contains the model's parameters.
        """
        # <!> shallow copy only
        self._parameters = parameters.copy()
        # convert lists
        for k, v in self._parameters.items():
            if isinstance(v, list):
                self._parameters[k] = np.array(v)



from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import Generic, Iterator, Tuple, TypeVar
from typing import Mapping as TMapping

__all__ = ["FilteredMappingProxy"]


KT = TypeVar("KT")
VT = TypeVar("VT")


@dataclass(frozen=True)
class FilteredMappingProxy(Mapping, Generic[KT, VT]):
    """
    Efficient filtered (with order) proxy (= no direct assignment) of a referenced mapping.

    Attributes
    ----------
    mapping : TMapping[KT, VT]
    subset : Tuple[KT, ...]
    check_keys : :obj:`bool`
    """

    mapping: TMapping[KT, VT]
    subset: Tuple[KT, ...]
    check_keys: bool = field(default=True, repr=False, compare=False)

    def __post_init__(self) -> None:
        # TODO? only store `types.MappingProxyType(self.mapping)` to be sure no write access?
        set_subset = set(self.subset)
        if len(set_subset) != len(self.subset):
            raise ValueError("Duplicate keys in the subset")
        if self.check_keys:
            unknown_keys = set_subset.difference(self.mapping)
            if len(unknown_keys):
                raise ValueError(f"Unknown keys in the subset: {unknown_keys}")

    def __len__(self) -> int:
        return len(self.subset)

    def __iter__(self) -> Iterator[KT]:
        return iter(self.subset)

    def __getitem__(self, k: KT) -> VT:
        if k not in self.subset:
            raise KeyError(k)
        return self.mapping[k]



"""This module defines the `OutputsSettings` and `AlgorithmSettings` classes."""

import json
import os
import shutil
import warnings
from pathlib import Path
from typing import Optional, Union

import torch

from leaspy.exceptions import LeaspyAlgoInputError
from leaspy.utils.typing import KwargsType

algo_default_data_dir = Path(__file__).parent.parent / "algo" / "data"

__all__ = [
    "OutputsSettings",
    "AlgorithmSettings",
    "algo_default_data_dir",
]


class OutputsSettings:
    """
    Used to create the `logs` folder to monitor the convergence of the fit algorithm.

    Parameters
    ----------
    settings : :obj:`dict` [:obj:`str`, Any]
            * path : :obj:`str` or None
                Path to store logs. If None, default path "./_outputs/" will be used.
            * print_periodicity : :obj:`int` >= 1 or None
                Print information every N iterations
            * save_periodicity : :obj:`int` >= 1, optional
                Save convergence data every N iterations
                Default=50.
            * plot_periodicity : :obj:`int` >= 1 or None
                Plot convergence data every N iterations.
                If None, no plots will be saved.
                Note that plotting requires saving to be realized and can not be more than saves.
            * plot_sourcewise : :obj:`bool`
                Flag to plot source based multidimensional parameters such as mixing_matrix for each source.
                Otherwise, they will be plotted according to the other dimension such as feature.
                Default=False
            * overwrite_logs_folder : :obj:`bool`
                Flag to remove all previous logs if existing (default False)

    Raises
    ------
    :exc:`.LeaspyAlgoInputError`
    """

    DEFAULT_LOGS_DIR = "_outputs"

    def __init__(self, settings):
        self.print_periodicity = None
        self.plot_periodicity = None
        self.save_periodicity = None
        self.plot_patient_periodicity = None
        self.plot_sourcewise = False
        self.nb_of_patients_to_plot = 5

        self.root_path = None
        self.parameter_convergence_path = None
        self.plot_path = None
        self.patients_plot_path = None

        self._set_print_periodicity(settings)
        self._set_save_periodicity(settings)
        self._set_plot_periodicity(settings)
        self._set_nb_of_patients_to_plot(settings)
        self._set_plot_sourcewise(settings)
        self._set_plot_patient_periodicity(settings)

        # only create folders if the user want to save data or plots and provided a valid path!
        self._create_root_folder(settings)

    def _set_param_as_int_or_ignore(self, settings: dict, param: str):
        """Inplace set of parameter (as int) from settings."""
        if param not in settings:
            return

        val = settings[param]
        if val is not None:
            # try to cast as an integer.
            try:
                val = int(val)
                assert val >= 1
            except Exception:
                warnings.warn(
                    f"The '{param}' parameter you provided is not castable to an int > 0. "
                    "Ignoring its value.",
                    UserWarning,
                )
                return

        # Update the attribute of self in-place
        setattr(self, param, val)

    def _set_plot_sourcewise(self, settings: dict):
        setattr(self, "plot_sourcewise", settings["plot_sourcewise"])

    def _set_nb_of_patients_to_plot(self, settings: dict):
        self._set_param_as_int_or_ignore(settings, "nb_of_patients_to_plot")

    def _set_print_periodicity(self, settings: dict):
        self._set_param_as_int_or_ignore(settings, "print_periodicity")

    def _set_save_periodicity(self, settings: dict):
        self._set_param_as_int_or_ignore(settings, "save_periodicity")

    def _set_plot_periodicity(self, settings: dict):
        self._set_param_as_int_or_ignore(settings, "plot_periodicity")

    def _set_plot_patient_periodicity(self, settings: dict):
        self._set_param_as_int_or_ignore(settings, "plot_patient_periodicity")

        if self.plot_periodicity is not None:
            if self.save_periodicity is None:
                raise LeaspyAlgoInputError(
                    "You can not define a `plot_periodicity` without defining `save_periodicity`. "
                    "Note that the `plot_periodicity` should be a multiple of `save_periodicity`."
                )

            if self.plot_periodicity % self.save_periodicity != 0:
                raise LeaspyAlgoInputError(
                    "The `plot_periodicity` should be a multiple of `save_periodicity`."
                )

    def _create_root_folder(self, settings: dict):
        # Get the path to put the outputs
        path = settings.get("path", None)
        if path is None:
            if self.save_periodicity:
                warnings.warn(
                    f"Outputs will be saved in '{self.DEFAULT_LOGS_DIR}' relative to the current working directory",
                    stacklevel=2,
                )
                path = Path.cwd() / self.DEFAULT_LOGS_DIR
                if path.exists():
                    self._clean_folder(path)
            else:
                return
        else:
            path = Path.cwd() / path

        settings["path"] = str(path)

        # Check if the folder does not exist: if not, create (and its parent)
        if not path.exists():
            warnings.warn(
                f"The logs path you provided ({settings['path']}) does not exist. "
                "Needed paths will be created (and their parents if needed).",
                stacklevel=2,
            )
        elif settings.get("overwrite_logs_folder", False):
            warnings.warn(f"Overwriting '{path}' folder...")
            self._clean_folder(path)

        all_ok = self._check_needed_folders_are_empty_or_create_them(path)

        if not all_ok:
            raise LeaspyAlgoInputError(
                f"The logs folder '{path}' already exists and is not empty! "
                "Give another path or use keyword argument `overwrite_logs_folder=True`."
            )

    @staticmethod
    def _check_folder_is_empty_or_create_it(path_folder: Path) -> bool:
        if path_folder.exists():
            if (
                os.path.islink(path_folder)
                or not path_folder.is_dir()
                or len([f for f in path_folder.iterdir()]) > 0
            ):
                return False
        else:
            path_folder.mkdir(parents=True, exist_ok=True)

        return True

    @staticmethod
    def _clean_folder(path: Path):
        shutil.rmtree(path)
        path.mkdir(exist_ok=True, parents=True)

    def _check_needed_folders_are_empty_or_create_them(self, path: Path) -> bool:
        self.root_path = path
        self.parameter_convergence_path = path / "parameter_convergence"
        self.plot_path = path / "plots"
        self.patients_plot_path = self.plot_path / "patients"
        all_ok = self._check_folder_is_empty_or_create_it(
            self.parameter_convergence_path
        )
        all_ok &= self._check_folder_is_empty_or_create_it(self.plot_path)
        all_ok &= self._check_folder_is_empty_or_create_it(self.patients_plot_path)

        return all_ok


class AlgorithmSettings:
    """
    Used to set the algorithms' settings.

    All parameters except the algorithm name have default values, which can be overwritten by the user.

    Parameters
    ----------
    name : str
        The algorithm's name. Must be one of:
            * For `fit` algorithms:
                * ``'mcmc_saem'``
                * ``'lme_fit'`` (for LME model only)
            * For `personalize` algorithms:
                * ``'scipy_minimize'``
                * ``'mean_real'``
                * ``'mode_real'``
                * ``'constant_prediction'`` (for constant model only)
                * ``'lme_personalize'`` (for LME model only)
            * For `simulate` algorithms:
                * ``'simulation'``

    **kwargs : any
        Depending on the algorithm, various parameters are possible:
            * seed : :obj:`int`, optional, default None
                Used for stochastic algorithms.
            * algorithm_initialization_method : :obj:`str`, optional
                Personalize the algorithm initialization method, according to those possible for the given algorithm
                (refer to its documentation in :mod:`leaspy.algo`).
            * n_iter : :obj:`int`, optional
                Number of iteration. Note that there is no stopping criteria for MCMC SAEM algorithms.
            * n_burn_in_iter : :obj:`int`, optional
                Number of iteration during burning phase, used for the MCMC SAEM algorithms.
            * use_jacobian : :obj:`bool`, optional, default True
                Used in ``scipy_minimize`` algorithm to perform a `L-BFGS` instead of a `Powell` algorithm.
            * n_jobs : :obj:`bool`, optional, default 1
                Used in ``scipy_minimize`` algorithm to accelerate calculation with parallel derivation using joblib.
            * progress_bar : :obj:`bool`, optional, default True
                Used to display a progress bar during computation.
            * device : :obj:`int` or torch.device, optional
                Specifies on which device the algorithm will run. Only 'cpu' and 'cuda' are supported for this argument.
                Only ``'mcmc_saem'``, ``'mean_real'`` and ``'mode_real'`` algorithms support this setting.

        For the complete list of the available parameters for a given algorithm, please directly refer to its documentation.

    Attributes
    ----------
    name : :obj:`str`
        The algorithm's name.
    algorithm_initialization_method : :obj:`str`, optional
        Personalize the algorithm initialization method, according to those possible for the given algorithm
        (refer to its documentation in :mod:`leaspy.algo`).
    seed : :obj:`int`, optional, default None
        Used for stochastic algorithms.
    parameters :  :obj:`dict`
        Contains the other parameters: `n_iter`, `n_burn_in_iter`, `use_jacobian`, `n_jobs` & `progress_bar`.
    logs : :class:`.OutputsSettings`, optional
        Used to create a ``logs`` file containing convergence information during fitting the model.
    device : :obj:`str` (or torch.device), optional, default 'cpu'
        Specifies the computation device. Only `'cpu'` and `'cuda'` are supported.
        Note that specifying an indexed CUDA device (such as 'cuda:1') is not supported.
        In order to specify the precise cuda device index, one should use the `CUDA_VISIBLE_DEVICES` environment variable.

    Raises
    ------
    :exc:`.LeaspyAlgoInputError`

    Notes
    -----
    Developers can use `_dynamic_default_parameters` to define settings that depend on other parameters when
    not explicitly specified by the user.

    """

    # TODO should be in the each algo class directly?
    _dynamic_default_parameters = {
        "lme_fit": [
            (
                lambda kw: "force_independent_random_effects" in kw
                and kw["force_independent_random_effects"],
                {
                    ("method",): lambda kw: [
                        "lbfgs",
                        "bfgs",
                    ]  # Powell & Nelder-Mead methods cannot ensure respect of "free"
                },
            )
        ]
    }

    # known keys for all algorithms (<!> not all of them are mandatory!)
    _known_keys = [
        "name",
        "seed",
        "algorithm_initialization_method",
        "parameters",
        "device",
    ]  # 'logs' are not handled in exported files

    def __init__(self, name: str, **kwargs):
        from leaspy.algo import AlgorithmName

        self.name: AlgorithmName = AlgorithmName(name)
        self.parameters: Optional[KwargsType] = None
        self.seed: Optional[int] = None
        self.algorithm_initialization_method: Optional[str] = None
        self.logs: Optional[OutputsSettings] = None
        default_algo_settings_path = (
            algo_default_data_dir / f"default_{self.name.value}.json"
        )
        if default_algo_settings_path.is_file():
            self._load_default_values(default_algo_settings_path)
        else:
            raise LeaspyAlgoInputError(
                f"The algorithm name '{self.name.value}' you provided does not exist"
            )
        self._manage_kwargs(kwargs)
        self.check_consistency()

    def check_consistency(self) -> None:
        """
        Check internal consistency of algorithm settings and warn or raise a `LeaspyAlgoInputError` if not.
        """
        from .algo_with_device import AlgorithmWithDeviceMixin
        from .base import get_algorithm_class

        algo_class = get_algorithm_class(self.name)
        if self.seed is not None and algo_class.deterministic:
            warnings.warn(
                f"You can skip defining `seed` since the algorithm {self.name} is deterministic."
            )
        if hasattr(self, "device") and not issubclass(
            algo_class, AlgorithmWithDeviceMixin
        ):
            warnings.warn(
                f'The algorithm "{self.name}" does not support user-specified devices (this '
                "is supported only for specific algorithms) and will use the default device (CPU)."
            )

    @classmethod
    def _recursive_merge_dict_warn_extra_keys(
        cls, ref: dict, new: dict, *, prefix_keys: str = ""
    ):
        """Merge in-place dictionary `ref` with the values from `new`, for dict keys, merge is recursive."""
        extra_keys = [prefix_keys + k for k in new if k not in ref]
        if extra_keys:
            warnings.warn(
                f"The parameters {extra_keys} were not present by default and are likely to be unsupported."
            )
        for k, v in new.items():
            if k not in ref or not isinstance(ref[k], dict):
                ref[k] = v
            else:
                if not isinstance(v, dict):
                    raise LeaspyAlgoInputError(
                        f"Algorithm parameter `{prefix_keys + k}` should be a dictionary, not '{v}' of type {type(v)}."
                    )
                cls._recursive_merge_dict_warn_extra_keys(
                    ref[k], v, prefix_keys=f"{prefix_keys}{k}."
                )

    @classmethod
    def load(cls, path_to_algorithm_settings: Union[str, Path]):
        """Instantiate a AlgorithmSettings object a from json file.

        Parameters
        ----------
        path_to_algorithm_settings :  :obj:`str`
            Path of the json file.

        Returns
        -------
        :class:`.AlgorithmSettings`
            An instanced of AlgorithmSettings with specified parameters.

        Raises
        ------
        :exc:`.LeaspyAlgoInputError`
            if anything is invalid in algo settings

        Examples
        --------
        >>> from leaspy.algo import AlgorithmSettings
        >>> leaspy_univariate = AlgorithmSettings.load('outputs/leaspy-univariate_model-settings.json')
        """
        with open(path_to_algorithm_settings) as fp:
            settings = json.load(fp)
        if "name" not in settings.keys():
            raise LeaspyAlgoInputError(
                "Your json file must contain a 'name' attribute!"
            )
        algorithm_settings = cls(settings["name"])
        if "parameters" in settings.keys():
            print("You overwrote the algorithm default parameters")
            cls._recursive_merge_dict_warn_extra_keys(
                algorithm_settings.parameters, cls._get_parameters(settings)
            )
        if "seed" in settings.keys():
            print("You overwrote the algorithm default seed")
            algorithm_settings.seed = cls._get_seed(settings)
        if "algorithm_initialization_method" in settings.keys():
            print("You overwrote the algorithm default initialization method")
            algorithm_settings.algorithm_initialization_method = (
                cls._get_algorithm_initialization_method(settings)
            )
        if "device" in settings.keys():
            print("You overwrote the algorithm default device")
            algorithm_settings.device = cls._get_device(settings)
        if "loss" in settings.keys():
            raise LeaspyAlgoInputError(
                "`loss` keyword for AlgorithmSettings is not supported any more. "
                "Please define `noise_model` directly in your Leaspy model."
            )
        # TODO: this class should really be refactored so not to copy in 3 methods same stuff (manage_kwargs, load & _check_default_settings)
        unknown_keys = set(settings.keys()).difference(cls._known_keys)
        if unknown_keys:
            raise LeaspyAlgoInputError(
                f"Unexpected keys {unknown_keys} in algorithm settings."
            )
        algorithm_settings.check_consistency()
        return algorithm_settings

    def save(self, path: Union[str, Path], **kwargs):
        """
        Save an AlgorithmSettings object in a json file.

        TODO? save leaspy version as well for retro/future-compatibility issues?

        Parameters
        ----------
        path : :obj:`str`
            Path to store the AlgorithmSettings.
        **kwargs
            Keyword arguments for json.dump method.
            Default: dict(indent=2)

        Examples
        --------
        >>> from leaspy.algo import AlgorithmSettings
        >>> settings = AlgorithmSettings("scipy_minimize", seed=42)
        >>> settings.save("outputs/scipy_minimize-settings.json")
        """
        from leaspy.algo import AlgorithmType, get_algorithm_type

        json_settings = {
            "name": self.name,
            "seed": self.seed,
            "algorithm_initialization_method": self.algorithm_initialization_method,
        }
        if hasattr(self, "device"):
            json_settings["device"] = self.device
        # TODO: save config of logging as well (OutputSettings needs to be JSON serializable...)
        # if self.logs is not None:
        #    json_settings['logs'] = self.logs
        # append parameters key after "hyperparameters"
        json_settings["parameters"] = self.parameters
        # Default json.dump kwargs:
        kwargs = {"indent": 2, **kwargs}
        with open(path, "w") as json_file:
            json.dump(json_settings, json_file, **kwargs)

    def set_logs(self, **kwargs):
        """
        Use this method to monitor the convergence of a model fit.

        This method creates CSV files and plots to track the evolution of population parameters
        (i.e., fixed effects) during the fitting.

        Parameters
        ----------
        **kwargs
            path : :obj:`str`, optional
               The path of the folder where graphs and csv files will be saved.
               If None, DEFAULT_LOGS_DIR will be used.
            * print_periodicity : :obj:`int`, optional, default 100
                Prints every N iterations.
            * save_periodicity : :obj:`int`, optional, default 50
                Saves the values in csv files every N iterations.
            * plot_periodicity : :obj:`int`, optional, default 1000
                Generates plots from saved values every N iterations.
                Notes:
                    * Must be a multiple of `save_periodicity`.
                    * Setting this value too low may significantly slow down the fitting process.
            * plot_patient_periodicity : :obj:`int`
                 Set the frequency of the saves of the patients' reconstructions
            * plot_sourcewise : :obj:`bool`, optional, default False
                Set this to True to plot the source-based parameters sourcewise.
            * overwrite_logs_folder : :obj:`bool`, optional, default False
                Set it to ``True`` to overwrite the content of the folder in ``path``.
            * nb_of_patients_to_plot : :obj:`int`, optional default 5
                number of patients to plot

        Raises
        ------
        :exc:`.LeaspyAlgoInputError`
            If the folder given in ``path`` already exists and if ``overwrite_logs_folder`` is set to ``False``.

        Notes
        -----
        By default, if the folder given in ``path`` already exists, the method will raise an error.
        To overwrite the content of the folder, set ``overwrite_logs_folder`` it to ``True``.
        """
        # TODO: all this logic should be delegated in dedicated OutputSettings class...!

        default_settings = {
            "path": None,
            "print_periodicity": None,
            "save_periodicity": None,
            "plot_periodicity": None,
            "plot_patient_periodicity": None,
            "plot_sourcewise": False,
            "overwrite_logs_folder": False,
            "nb_of_patients_to_plot": 5,
        }
        settings = default_settings.copy()
        for k, v in kwargs.items():
            if k in (
                "print_periodicity",
                "plot_periodicity",
                "save_periodicity",
                "plot_patient_periodicity",
                "nb_of_patients_to_plot",
                "plot_sourcewise",
            ):
                if v is not None and not isinstance(v, int):
                    raise LeaspyAlgoInputError(
                        f"You must provide a integer to the input <{k}>! "
                        f"You provide {v} of type {type(v)}."
                    )
                settings[k] = v
            elif k in ["overwrite_logs_folder"]:
                if not isinstance(v, bool):
                    raise LeaspyAlgoInputError(
                        f"You must provide a boolean to the input <{k}>! "
                        f"You provide {v} of type {type(v)}."
                    )
                settings[k] = v
            elif k == "path":
                if v is not None and not isinstance(v, (str, Path)):
                    raise LeaspyAlgoInputError(
                        f"You must provide a string or Path to the input <{k}>! "
                        f"You provide {v} of type {type(v)}."
                    )
                settings[k] = v
        if settings != default_settings:
            self.logs = OutputsSettings(settings)

    def _manage_kwargs(self, kwargs):
        _special_kwargs = {
            "seed": self._get_seed,
            "algorithm_initialization_method": self._get_algorithm_initialization_method,
            "device": self._get_device,
        }

        for k, v in kwargs.items():
            if k in _special_kwargs:
                k_getter = _special_kwargs[k]
                setattr(self, k, k_getter(kwargs))

        kwargs_interpreted_as_parameters = {
            k: v for k, v in kwargs.items() if k not in _special_kwargs
        }
        self._recursive_merge_dict_warn_extra_keys(
            self.parameters, kwargs_interpreted_as_parameters
        )

        # dynamic default parameters
        if self.name in self._dynamic_default_parameters:
            for func_condition, associated_defaults in self._dynamic_default_parameters[
                self.name
            ]:
                if not func_condition(kwargs):
                    continue

                # loop on dynamic defaults
                for nested_levels, val_getter in associated_defaults.items():
                    # check that the dynamic default that we want to set is not already overwritten
                    if self._get_nested_dict(kwargs, nested_levels) is None:
                        self._set_nested_dict(
                            self.parameters, nested_levels, val_getter(kwargs)
                        )

    @staticmethod
    def _get_nested_dict(nested_dict: dict, nested_levels, default=None):
        """
        Get a nested key of a dict or default if any previous level is missing.

        Examples
        --------
        >>> _get_nested_dict(d, ('a','b'), -1) == ...
            * -1 if 'a' not in d
            * -1 if 'b' not in d['a']
            * d['a']['b'] else

        >>> _get_nested_dict(d, (), ...) == d
        """
        it_levels = iter(nested_levels)

        while isinstance(nested_dict, dict):
            try:
                next_lvl = next(it_levels)
            except StopIteration:
                break

            # get next level dict
            nested_dict = nested_dict.get(next_lvl, default)

        return nested_dict

    @classmethod
    def _set_nested_dict(cls, nested_dict: dict, nested_levels, val):
        """
        Set a nested key of a dict.
        Precondition: all intermediate levels must exist.
        """
        *nested_top_levels, last_level = nested_levels
        dict_to_set = cls._get_nested_dict(nested_dict, nested_top_levels, default=None)
        assert isinstance(dict_to_set, dict)
        dict_to_set[last_level] = val  # inplace

    def _load_default_values(self, path_to_algorithm_settings: Path):
        from leaspy.algo import AlgorithmType, get_algorithm_class, get_algorithm_type

        with open(path_to_algorithm_settings) as fp:
            settings = json.load(fp)
        self._check_default_settings(settings)
        # TODO: Urgent => The following function should in fact be algorithm-name specific!! As for the constant prediction
        # Etienne: I'd advocate for putting all non-generic / parametric stuff in special methods / attributes
        # of corresponding algos... so that everything is generic here
        # Igor : Agreed. This class became a real mess.
        self.name = self._get_name(settings)
        self.parameters = self._get_parameters(settings)
        self.algorithm_initialization_method = (
            self._get_algorithm_initialization_method(settings)
        )
        # optional hyperparameters depending on type of algorithm
        algo_class = get_algorithm_class(self.name)
        if not algo_class.deterministic:
            self.seed = self._get_seed(settings)
        if "device" in settings:
            self.device = self._get_device(settings)

    @classmethod
    def _check_default_settings(cls, settings: dict):
        from leaspy.algo import AlgorithmType, get_algorithm_class, get_algorithm_type

        unknown_keys = set(settings.keys()).difference(cls._known_keys)
        if unknown_keys:
            raise LeaspyAlgoInputError(
                f"Unexpected keys {unknown_keys} in algorithm settings."
            )
        error_tpl = "The '{}' key is missing in the algorithm settings (JSON file) you are loading."
        for mandatory_key in ("name", "parameters"):
            if mandatory_key not in settings.keys():
                raise LeaspyAlgoInputError(error_tpl.format(mandatory_key))
        algo_class = get_algorithm_class(settings["name"])
        if not algo_class.deterministic and "seed" not in settings:
            raise LeaspyAlgoInputError(error_tpl.format("seed"))
        if "algorithm_initialization_method" not in settings:
            raise LeaspyAlgoInputError(
                error_tpl.format("algorithm_initialization_method")
            )

    @staticmethod
    def _get_name(settings: dict) -> str:
        return settings["name"].lower()

    @staticmethod
    def _get_parameters(settings: dict) -> dict:
        return settings["parameters"]

    @staticmethod
    def _get_seed(settings: dict) -> Optional[int]:
        if settings["seed"] is None:
            return None
        try:
            return int(settings["seed"])
        except Exception:
            warnings.warn(
                f"The 'seed' parameter you provided ({settings['seed']}) cannot be converted to int, using None instead."
            )
            return None

    @staticmethod
    def _get_algorithm_initialization_method(settings: dict) -> Optional[str]:
        if settings["algorithm_initialization_method"] is None:
            return None
        # TODO : There should be a list of possible initialization method.
        #  It can also be discussed depending on the algorithms name
        return settings["algorithm_initialization_method"]

    @staticmethod
    def _get_device(settings: dict):
        # in case where a torch.device object was used, we convert it to the
        # corresponding string (torch.device('cuda') is converted into 'cuda')
        # in order for the AlgorithmSettings to be saved into json files if needed
        if isinstance(settings["device"], torch.device):
            return settings["device"].type

        # getting the type of torch.device(...) allows to convert 'cuda:2' to 'cuda'
        # which prevents potential issues when using torch.set_default_tensor_type
        return torch.device(settings["device"]).type



from .base import BaseModel, ModelInterface
from .constant import ConstantModel
from .factory import ModelName, model_factory
from .joint import JointModel
from .linear import LinearModel
from .lme import LMEModel
from .logistic import LogisticModel
from .mcmc_saem_compatible import McmcSaemCompatibleModel
from .riemanian_manifold import RiemanianManifoldModel
from .settings import ModelSettings
from .shared_speed_logistic import SharedSpeedLogisticModel
from .stateful import StatefulModel
from .stateless import StatelessModel
from .time_reparametrized import TimeReparametrizedModel
from .mixture import LogisticMultivariateMixtureModel

__all__ = [
    "ModelInterface",
    "ModelName",
    "McmcSaemCompatibleModel",
    "TimeReparametrizedModel",
    "BaseModel",
    "ConstantModel",
    "StatelessModel",
    "StatefulModel",
    "LMEModel",
    "model_factory",
    "ModelSettings",
    "RiemanianManifoldModel",
    "LogisticModel",
    "LinearModel",
    "SharedSpeedLogisticModel",
    "JointModel",
    "LogisticMultivariateMixtureModel",
]



from .abstract_attributes import AbstractAttributes
from .linear_attributes import LinearAttributes
from .logistic_attributes import LogisticAttributes
from .logistic_parallel_attributes import LogisticParallelAttributes

__all__ = [
    "AttributesFactory",
    "LinearAttributes",
    "LogisticAttributes",
    "LogisticParallelAttributes",
]


class AttributesFactory:
    """
    Return an `Attributes` class object based on the given parameters.
    """

    _attributes = {
        "logistic": LogisticAttributes,
        "univariate_logistic": LogisticAttributes,
        "logistic_parallel": LogisticParallelAttributes,
        "linear": LinearAttributes,
        "univariate_linear": LinearAttributes,
        #'mixed_linear-logistic': ... # TODO
    }

    @classmethod
    def attributes(
        cls, name: str, dimension: int, source_dimension: int = None
    ) -> AbstractAttributes:
        """
        Class method to build correct model attributes depending on model `name`.

        Parameters
        ----------
        name : str
        dimension : int
        source_dimension : int, optional (default None)

        Returns
        -------
        :class:`.AbstractAttributes`

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            if any inconsistent parameter.
        """
        from leaspy.exceptions import LeaspyModelInputError

        if isinstance(name, str):
            name = name.lower()
        else:
            raise LeaspyModelInputError("The `name` argument must be a string!")

        if name not in cls._attributes:
            raise LeaspyModelInputError(
                f"The name '{name}' you provided for the attributes is not supported."
                f"Valid choices are: {list(cls._attributes.keys())}"
            )

        if not (("univariate" in name) ^ (dimension != 1)):
            raise LeaspyModelInputError(
                f"Name `{name}` should contain 'univariate', if and only if `dimension` equals 1."
            )

        return cls._attributes[name](name, dimension, source_dimension)



from .base import PersonalizeAlgorithm
from .constant_prediction_algo import ConstantPredictionAlgorithm
from .lme_personalize import LMEPersonalizeAlgorithm
from .mcmc import McmcPersonalizeAlgorithm
from .mean_posterior import MeanPosteriorAlgorithm
from .mode_posterior import ModePosteriorAlgorithm
from .scipy_minimize import ScipyMinimizeAlgorithm

__all__ = [
    "McmcPersonalizeAlgorithm",
    "PersonalizeAlgorithm",
    "MeanPosteriorAlgorithm",
    "ModePosteriorAlgorithm",
    "ScipyMinimizeAlgorithm",
    "ConstantPredictionAlgorithm",
    "LMEPersonalizeAlgorithm",
]



"""This module defines commonly used named input functions."""

from __future__ import annotations

import torch

from ..linalg import compute_orthonormal_basis
from ..weighted_tensor import factory_weighted_tensor_unary_operator, sum_dim
from ._named_input_function import NamedInputFunction
from ._utils import _arguments_checker, _identity, _prod_args, _sum_args

__all__ = [
    "Prod",
    "Identity",
    "MatMul",
    "OrthoBasis",
    "Exp",
    "Sqr",
    "Mean",
    "Std",
    "SumDim",
    "Sum",
]


Prod = NamedInputFunction.bound_to(
    _prod_args,
    _arguments_checker(
        possible_kws={"start"},
    ),
)


Identity = NamedInputFunction.bound_to(
    _identity,
    _arguments_checker(
        nb_arguments=1,
        possible_kws=set(),
    ),
)


MatMul = NamedInputFunction.bound_to(
    torch.matmul,
    _arguments_checker(
        nb_arguments=2,
        possible_kws=set(),
    ),
)


OrthoBasis = NamedInputFunction.bound_to(
    compute_orthonormal_basis,
    _arguments_checker(
        nb_arguments=2,
        possible_kws=set("strip_col"),
    ),
)


Exp = NamedInputFunction.bound_to(
    factory_weighted_tensor_unary_operator(torch.exp),
    _arguments_checker(
        nb_arguments=1,
        possible_kws=set(),
    ),
)


Sqr = NamedInputFunction.bound_to(
    factory_weighted_tensor_unary_operator(torch.square),
    _arguments_checker(
        nb_arguments=1,
        possible_kws=set(),
    ),
)


Mean = NamedInputFunction.bound_to(
    # <!> never compute mean directly on WeightedTensor (use `.wsum()` instead)
    torch.mean,
    _arguments_checker(
        nb_arguments=1,
        possible_kws={"dim"},
    ),
)


Std = NamedInputFunction.bound_to(
    # <!> never compute std directly on WeightedTensor
    torch.std,
    _arguments_checker(
        nb_arguments=1,
        possible_kws={"dim", "unbiased"},
    ),
)


SumDim = NamedInputFunction.bound_to(
    sum_dim,
    _arguments_checker(
        nb_arguments=1,  # with `dim` XOR `but_dim`
    ),
)


Sum = NamedInputFunction.bound_to(
    _sum_args,
    _arguments_checker(
        possible_kws={"start"},
    ),
)


# Filled = NamedInputFunction.bound_to(
#     WeightedTensor.filled, arguments_checker(nb_arguments=1, mandatory_kws={"fill_value"})
# )
# Negate = NamedInputFunction.bound_to(
#     operator.neg, arguments_checker(nb_arguments=1, possible_kws=set())
# )
# ItemGetter = NamedInputFunction.bound_to(
#     operator.itemgetter, arguments_checker(nb_arguments=1, possible_kws=set())
# )



from .loader import (
    DatasetName,
    get_dataset_path,
    get_individual_parameter_path,
    get_model_path,
    load_dataset,
    load_individual_parameters,
    load_model,
)

__all__ = [
    "DatasetName",
    "load_dataset",
    "load_individual_parameters",
    "load_model",
    "get_dataset_path",
    "get_model_path",
    "get_individual_parameter_path",
]



import pandas as pd
import torch

from leaspy.io.data.dataset import Dataset
from leaspy.utils.functional import Exp, OrthoBasis, Sqr
from leaspy.utils.weighted_tensor import (
    unsqueeze_right,
)
from leaspy.variables.distributions import Normal
from leaspy.variables.specs import (
    Hyperparameter,
    ModelParameter,
    NamedVariables,
    PopulationLatentVariable,
    VariableNameToValueMapping,
)

from .obs_models import FullGaussianObservationModel
from .riemanian_manifold import RiemanianManifoldModel

__all__ = [
    "LinearInitializationMixin",
    "LinearModel",
]


class LinearInitializationMixin:
    """Compute initial values for model parameters."""

    def _compute_initial_values_for_model_parameters(
        self,
        dataset: Dataset,
    ) -> VariableNameToValueMapping:
        """
        Compute and return initial values for the model parameters using linear regression.

        Parameters
        ----------
        dataset : :class:`Dataset`
            The dataset from which to extract observations and masks.

        Returns
        -------
        :class:`~leaspy.variables.specs.VariableNameToValueMapping`
            A dictionary mapping parameter names (as strings) to their initialized
            torch.Tensor values.

        Notes
        -----
        - The initial positions are computed using `intercept + t0 * slope` where `t0`
        is the mean of all observation times.
        - Velocities are averaged per feature, and log-transformed using
        `get_log_velocities`.
        - Time parameters (`tau_mean`, `tau_std`) and variability parameters
        (`xi_std`) are also initialized.
        - If `self.source_dimension >= 1`, zero initialization is used for `betas_mean`.
        - If the observation model is a `FullGaussianObservationModel`, the
        `noise_std` parameter is also added, expanded to the correct shape.
        """
        from leaspy.models.utilities import (
            compute_linear_regression_subjects,
            get_log_velocities,
            torch_round,
        )

        df = dataset.to_pandas(apply_headers=True)
        times = df.index.get_level_values("TIME").values
        t0 = times.mean()

        d_regress_params = compute_linear_regression_subjects(df, max_inds=None)
        df_all_regress_params = pd.concat(d_regress_params, names=["feature"])
        df_all_regress_params["position"] = (
            df_all_regress_params["intercept"] + t0 * df_all_regress_params["slope"]
        )
        df_grp = df_all_regress_params.groupby("feature", sort=False)
        positions = torch.tensor(df_grp["position"].mean().values)
        velocities = torch.tensor(df_grp["slope"].mean().values)

        parameters = {
            "g_mean": positions,
            "log_v0_mean": get_log_velocities(velocities, self.features),
            "tau_mean": torch.tensor(t0),
            "tau_std": self.tau_std,
            "xi_std": self.xi_std,
        }
        if self.source_dimension >= 1:
            parameters["betas_mean"] = torch.zeros(
                (self.dimension - 1, self.source_dimension)
            )
        rounded_parameters = {
            str(p): torch_round(v.to(torch.float32)) for p, v in parameters.items()
        }
        obs_model = next(iter(self.obs_models))  # WIP: multiple obs models...
        if isinstance(obs_model, FullGaussianObservationModel):
            rounded_parameters["noise_std"] = self.noise_std.expand(
                obs_model.extra_vars["noise_std"].shape
            )
        return rounded_parameters


class LinearModel(LinearInitializationMixin, RiemanianManifoldModel):
    """Manifold model for multiple variables of interest (linear formulation)."""

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)

    def get_variables_specs(self) -> NamedVariables:
        """
        Return the specifications of the variables (latent variables, derived variables,
        model 'parameters') that are part of the model.

        Returns
        -------
        :class:`~leaspy.variables.specs.NamedVariables :
            A dictionary-like object mapping variable names to their specifications.
        """
        d = super().get_variables_specs()
        d.update(
            g_mean=ModelParameter.for_pop_mean("g", shape=(self.dimension,)),
            g_std=Hyperparameter(0.01),
            g=PopulationLatentVariable(Normal("g_mean", "g_std")),
        )

        return d

    @staticmethod
    def metric(*, g: torch.Tensor) -> torch.Tensor:
        """
        Compute the metric tensor for the model.

        Parameters
        ----------
        g :  :class:`torch.Tensor`
            Input tensor with values of the population parameter `g` for each feature.

        Returns
        -------
         :class:`torch.Tensor`
            A tensor of ones with the same shape as `g`.
        """
        return torch.ones_like(g)

    @classmethod
    def model_with_sources(
        cls,
        *,
        rt: torch.Tensor,
        space_shifts: torch.Tensor,
        metric,
        v0,
        g,
    ) -> torch.Tensor:
        """
        Return the model output when sources(spatial components) are present.

        Parameters
        ----------
        rt :  :class:`torch.Tensor`
            The reparametrized time.
        space_shifts :  :class:`torch.Tensor`
            The values of the space-shifts
        metric : Any
            The metric tensor used for computing the spatial/temporal influence.
        v0 : Any
            The values of the population parameter `v0` for each feature.
        g : Any
            The values of the population parameter `g` for each feature.

        Returns
        -------
         :class:`torch.Tensor`
            The model output with contribution from sources.
        """
        pop_s = (None, None, ...)
        rt = unsqueeze_right(rt, ndim=1)  # .filled(float('nan'))
        return (g[pop_s] + v0[pop_s] * rt + space_shifts[:, None, ...]).weighted_value



import functools
import json
import operator
import os
import warnings
from typing import Callable, Iterable

import numpy as np
import pandas as pd
import torch

from leaspy.exceptions import LeaspyIndividualParamsInputError
from leaspy.utils.typing import DictParams, DictParamsTorch, IDType, ParamType

__all__ = ["IndividualParameters"]


class IndividualParameters:
    r"""
    Data container for individual parameters, contains IDs, timepoints and observations values.
    Output of the :class:`.leaspy.algo.personalize` method, contains the *random effects*.

    There are used as output of the `personalization algorithms` and as input/output of the `simulation algorithm`,
    to provide an initial distribution of individual parameters.

    Attributes
    ----------
    _indices : :obj:`list`
        List of the patient indices
    _individual_parameters : :obj:`dict`
        Individual indices (key) with their corresponding individual parameters {parameter name: parameter value}
    _parameters_shape : :obj:`dict`
        Shape of each individual parameter
    _default_saving_type : :obj:`str`
        Default extension for saving when none is provided
    """

    VALID_IO_EXTENSIONS = ["csv", "json"]

    def __init__(self):
        self._indices: list[IDType] = []
        self._individual_parameters: dict[IDType, DictParams] = {}
        self._parameters_shape = None  # {p_name: p_shape as tuple}
        self._default_saving_type = "csv"

    @property
    def _parameters_size(self) -> dict[ParamType, int]:
        """
        Get the size (number of scalar elements) of each parameter from its shape.
        It converts each parameter's shape into its flat size.

        Returns
        -------
       :obj:`dict` of ParamType to :obj:`int`
            A dictionary mapping each parameter type to its total number of scalar values.

        Examples
        --------
            * shape ``()`` becomes size ``1```
            * shape ``(1,)``becomes size ``1```
            * shape ``(2,3)``becomes size ``6```

        """
        shape_to_size = lambda shape: functools.reduce(operator.mul, shape, 1)

        return {p: shape_to_size(s) for p, s in self._parameters_shape.items()}

    def add_individual_parameters(
        self, index: IDType, individual_parameters: DictParams
    ):
        r"""
        Add the individual parameter of an individual to the IndividualParameters object

        Parameters
        ----------
        index : :class::class:`~leaspy..utils.typing.IDType`
            Index of the individual
        individual_parameters : :class:`~leaspy.utils.typing.DictParams`
            Individual parameters of the individual

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            * If the index is not a string or has already been added
            * Or if the individual parameters is not a dict.
            * Or if individual parameters are not self-consistent.

        Examples
        --------
        Add two individual with tau, xi and sources parameters

        >>> ip = IndividualParameters()
        >>> ip.add_individual_parameters('index-1', {"xi": 0.1, "tau": 70, "sources": [0.1, -0.3]})
        >>> ip.add_individual_parameters('index-2', {"xi": 0.2, "tau": 73, "sources": [-0.4, -0.1]})
        """
        # Check indices
        if not isinstance(index, str):
            raise LeaspyIndividualParamsInputError(
                f"The index should be a string ({type(index)} provided instead)"
            )

        if index in self._indices:
            raise LeaspyIndividualParamsInputError(
                f"The index {index} has already been added before"
            )

        # Check the dictionary format
        if not isinstance(individual_parameters, dict):
            raise LeaspyIndividualParamsInputError(
                "The `individual_parameters` argument should be a dictionary"
            )

        # Conversion of numpy arrays to lists
        individual_parameters = {
            k: v.tolist() if isinstance(v, np.ndarray) else v
            for k, v in individual_parameters.items()
        }

        # Check types of params
        for k, v in individual_parameters.items():
            valid_scalar_types = [
                int,
                np.int32,
                np.int64,
                float,
                np.float32,
                np.float64,
            ]

            scalar_type = type(v)
            if isinstance(v, list):
                scalar_type = None if len(v) == 0 else type(v[0])
            # elif isinstance(v, np.ndarray):
            #    scalar_type = v.dtype

            if scalar_type not in valid_scalar_types:
                raise LeaspyIndividualParamsInputError(
                    f"Incorrect dictionary value. Error for key: {k} -> scalar type {scalar_type}"
                )

        # Fix/check parameters nomenclature and shapes
        # (scalar or 1D arrays only...)
        pshapes = {
            p: (len(v),) if isinstance(v, list) else ()
            for p, v in individual_parameters.items()
        }

        if self._parameters_shape is None:
            # Keep track of the parameter shape
            self._parameters_shape = pshapes
        elif self._parameters_shape != pshapes:
            raise LeaspyIndividualParamsInputError(
                f"Invalid parameter shapes provided: {pshapes}. Expected: {self._parameters_shape}. "
                "Some parameters may be missing/unknown or have a wrong shape."
            )

        # Finally: add to internal dict object + indices array
        self._indices.append(index)
        self._individual_parameters[index] = individual_parameters

    def __getitem__(self, item: IDType) -> DictParams:
        """
        Get the individual parameters for a given individual.

        Parameters
        ----------
        item : :class:`~leaspy.utils.typing.IDType`

        Returns
        -------
        :class:`~leaspy.utils.typing.DictParams
            A dictionary containing the individual parameters for the specific individual.

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            * If the provided 'item' is not IDType (string)
            * If no individual with this ID has been found
        """
        if not isinstance(item, IDType):
            raise LeaspyIndividualParamsInputError(
                f"The index should be a string ({type(item)} provided instead)"
            )
        if item not in self._individual_parameters:
            raise LeaspyIndividualParamsInputError(f"The index {item} is unknown")
        return self._individual_parameters[item]

    def items(self):
        """
        Get the items of the individual parameters dictionary.

        Returns
        -------
        ItemsView
            A view object displaying a list of tuples (individual ID, parameters dict)
        
        Examples
        --------
        >>> ip = IndividualParameters()
        >>> ip.add_individual_parameters('index-1', {"xi": 0.1, "tau": 70, "sources": [0.1, -0.3]})
        >>> list(ip.items())
        ['index-1', {"xi": 0.1, "tau": 70, "sources": [0.1, -0.3]}]
        """
        return self._individual_parameters.items()

    def subset(self, indices: Iterable[IDType], *, copy: bool = True):
        r"""
        Returns IndividualParameters object with a subset of the initial individuals

        Parameters
        ----------
        indices : :obj:`Iterable`[:class:`~leaspy.utils.typing.IDType`]
            List of strings that corresponds to the indices of the individuals to return
        copy : :obj:`bool`, optional (default True)
            Should we copy underlying parameters or not?

        Returns
        -------
        :class:`.IndividualParameters`
            An instance of the IndividualParameters object with the selected list of individuals

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            Raise an error if one of the index is not in the IndividualParameters

        Examples
        --------
        >>> ip = IndividualParameters()
        >>> ip.add_individual_parameters('index-1', {"xi": 0.1, "tau": 70, "sources": [0.1, -0.3]})
        >>> ip.add_individual_parameters('index-2', {"xi": 0.2, "tau": 73, "sources": [-0.4, -0.1]})
        >>> ip.add_individual_parameters('index-3', {"xi": 0.3, "tau": 58, "sources": [-0.6, 0.2]})
        >>> ip_sub = ip.subset(['index-1', 'index-3'])
        """
        ip = IndividualParameters()

        unknown_ix = [ix for ix in indices if ix not in self._indices]
        if len(unknown_ix) > 0:
            raise LeaspyIndividualParamsInputError(
                f"The index {unknown_ix} are not in the indices."
            )

        for idx in indices:
            p = self[idx]
            if copy:
                p = p.copy()  # deepcopy here?
            ip.add_individual_parameters(idx, p)

        return ip

    def get_aggregate(self, parameter: ParamType, function: Callable) -> list:
        r"""
        Returns the result of aggregation by `function` of parameter values across all patients

        Parameters
        ----------
        parameter : :class:`~leaspy..utils.typing.ParamType`
            Name of the parameter
        function : :obj:`Callable`
            A function operating on iterables and supporting axis keyword,
            and outputing an iterable supporting the `tolist` method.

        Returns
        -------
        :obj:`list`
            Resulting value of the parameter

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            * If individual parameters are empty,
            * or if the parameter is not in the IndividualParameters.

        Examples
        --------
        >>> ip = IndividualParameters.load("path/to/individual_parameters")
        >>> tau_median = ip.get_aggregate("tau", np.median)
        """
        if self._parameters_shape is None:
            raise LeaspyIndividualParamsInputError(
                f"Individual parameters are empty: no information on '{parameter}'."
            )
        if parameter not in self._parameters_shape.keys():
            raise LeaspyIndividualParamsInputError(
                f"Parameter '{parameter}' does not exist in the individual parameters"
            )

        p = [v[parameter] for v in self._individual_parameters.values()]
        p_agg = function(p, axis=0).tolist()

        return p_agg

    def get_mean(self, parameter: ParamType):
        r"""
        Returns the mean value of a parameter across all patients

        Parameters
        ----------
        parameter : :class:`~leaspy.utils.typing.ParamType`
            Name of the parameter

        Returns
        -------
        :obj:`list`
            Mean value of the parameter

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            * If individual parameters are empty,
            * or if the parameter is not in the IndividualParameters.

        Examples
        --------
        >>> ip = IndividualParameters.load("path/to/individual_parameters")
        >>> tau_mean = ip.get_mean("tau")
        """
        return self.get_aggregate(parameter, np.mean)

    def get_std(self, parameter: ParamType):
        r"""
        Returns the standard deviation of a parameter across all patients

        Parameters
        ----------
        parameter : :class:`~leaspy.utils.typing.ParamType`
            Name of the parameter

        Returns
        -------
        :obj:`list`
            Standard-deviation value of the parameter

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            * If individual parameters are empty,
            * or if the parameter is not in the IndividualParameters.

        Examples
        --------
        >>> ip = IndividualParameters.load("path/to/individual_parameters")
        >>> tau_std = ip.get_std("tau")
        """
        return self.get_aggregate(parameter, np.std)

    def to_dataframe(self) -> pd.DataFrame:
        r"""
        Returns the dataframe of individual parameters

        Returns
        -------
        :class:`pandas.DataFrame`
            Each row corresponds to one individual.
            The index corresponds to the individual index ('ID').
            The columns are the names of the parameters.

        Examples
        --------
        Convert the individual parameters object into a dataframe

        >>> ip = IndividualParameters.load("path/to/individual_parameters")
        >>> ip_df = ip.to_dataframe()
        """
        # Get the data, idx per idx
        arr = []
        for idx in self._indices:
            indiv_arr = [idx]
            indiv_p = self._individual_parameters[idx]

            for p_name, p_shape in self._parameters_shape.items():
                if p_shape == ():
                    indiv_arr.append(indiv_p[p_name])
                else:
                    indiv_arr += indiv_p[p_name]  # 1D array only...
            arr.append(indiv_arr)

        # Get the column names
        final_names = ["ID"]
        for p_name, p_shape in self._parameters_shape.items():
            if p_shape == (1,) and "source" not in p_name:
                final_names.append(p_name)
            else:
                final_names += [
                    p_name + "_" + str(i) for i in range(p_shape[0])
                ]  # 1D array only...

        df = pd.DataFrame(arr, columns=final_names)
        return df.set_index("ID")

    @staticmethod
    def from_dataframe(df: pd.DataFrame):
        r"""
        Static method that returns an IndividualParameters object from the dataframe

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Dataframe of the individual parameters. Each row must correspond to one individual. The index corresponds
            to the individual index. The columns are the names of the parameters.

        Returns
        -------
        :class:`.IndividualParameters`
            An instance of IndividualParameters initialized from the DataFrame.

        Examples
        --------
        >>> import pandas as pd
        >>> data = {
        >>>     'tau': [70, 73],
        >>>     'xi': [0.1, 0.2],
        >>>     'sources_0': [0.1, -0.4],
        >>>     'sources_1': [-0.3, -0.1]
        >>> }
        >>> df = pd.DataFrame(data, index=['id1', 'id2'])
        >>> ip = IndividualParameters.from_dataframe(df)
        """
        # Check the names to keep
        df_names: list[ParamType] = list(df.columns.values)

        final_names = {}
        for name in df_names:
            split = name.split("_")[0]
            if split == name:  # e.g tau, xi, ...
                final_names[name] = name
            else:  # e.g sources_0 --> sources
                if split not in final_names:
                    final_names[split] = []
                final_names[split].append(name)

        # Create the individual parameters
        ip = IndividualParameters()

        for idx, row in df.iterrows():
            i_d = {
                param: np.array(row[col].tolist())
                if isinstance(col, list)
                else np.array([row[col]])
                for param, col in final_names.items()
            }
            ip.add_individual_parameters(idx, i_d)

        return ip

    @staticmethod
    def from_pytorch(indices: list[IDType], dict_pytorch: DictParamsTorch):
        r"""
        Static method that returns an IndividualParameters object from the indices and pytorch dictionary

        Parameters
        ----------
        indices : :obj:`list`[:class:`~leaspy.utils.typing.IDType`]
            List of the patients indices
        dict_pytorch : :class:`~leaspy.utils.typing.DictParmasTorch`
            Dictionary of the individual parameters

        Returns
        -------
        :class:`.IndividualParameters`
            An instance of IndividualParameters initialized from the pytorch dictionary.

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`

        Examples
        --------
        >>> indices = ['index-1', 'index-2', 'index-3']
        >>> ip_pytorch = {
        >>>    "xi": torch.tensor([[0.1], [0.2], [0.3]], dtype=torch.float32),
        >>>    "tau": torch.tensor([[70], [73], [58.]], dtype=torch.float32),
        >>>    "sources": torch.tensor([[0.1, -0.3], [-0.4, 0.1], [-0.6, 0.2]], dtype=torch.float32)
        >>> }
        >>> ip_pytorch = IndividualParameters.from_pytorch(indices, ip_pytorch)
        """

        len_p = {k: len(v) for k, v in dict_pytorch.items()}
        for k, v in len_p.items():
            if v != len(indices):
                raise LeaspyIndividualParamsInputError(
                    f"The parameter {k} should be of same length as the indices"
                )

        ip = IndividualParameters()

        keys = list(dict_pytorch.keys())

        for i, idx in enumerate(indices):
            p = {k: dict_pytorch[k][i].tolist() for k in keys}

            ip.add_individual_parameters(idx, p)

        return ip

    def to_pytorch(self) -> tuple[list[IDType], DictParamsTorch]:
        r"""
        Returns the indices and pytorch dictionary of individual parameters

        Returns
        -------
        indices: :obj:`list`[:class:`~leaspy.utils.typing.IDType`]
            List of patient indices
        pytorch_dict: :class:`~leaspy.utils.typing.DictParamsTorch`
            Dictionary of the individual parameters {parameter name: pytorch tensor of values across individuals}

        Examples
        --------
        Convert the individual parameters object into a dataframe

        >>> ip = IndividualParameters.load("path/to/individual_parameters")
        >>> indices, ip_pytorch = ip.to_pytorch()
        """
        ips_pytorch = {}

        for p_name, p_size in self._parameters_size.items():
            p_val = [self._individual_parameters[idx][p_name] for idx in self._indices]
            p_val = torch.tensor(p_val, dtype=torch.float32)
            p_val = p_val.reshape(shape=(len(self._indices), p_size))  # always 2D

            ips_pytorch[p_name] = p_val

        return self._indices, ips_pytorch

    def save(self, path: str, **kwargs):
        r"""
        Saves the individual parameters (json or csv) at the path location

        TODO? save leaspy version as well for retro/future-compatibility issues?

        Parameters
        ----------
        path : :obj:`str`
            Path and file name of the individual parameters. The extension can be json or csv.
            If no extension, default extension (csv) is used
        **kwargs
            Additional keyword arguments to pass to either:
            * :meth:`pandas.DataFrame.to_csv`
            * :func:`json.dump`
            depending on saving format requested

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            * If extension not supported for saving
            * If individual parameters are empty
        
        Warnings
        --------
        Emits a warning if no file extension is provided and the default extension is used.

        Examples
        --------
        >>> ip.save("params.csv", index=False)
        >>> ip.save("params.json", indent=4)
        """
        if self._parameters_shape is None:
            raise LeaspyIndividualParamsInputError(
                "Individual parameters are empty: unable to save them."
            )

        extension = self._check_and_get_extension(path)
        if extension is None:
            warnings.warn(
                f"You did not provide a valid extension (csv or json) for the file. "
                f"Default to {self._default_saving_type}."
            )
            extension = self._default_saving_type
            path = path + "." + extension

        if extension == "csv":
            self._save_csv(path, **kwargs)
        elif extension == "json":
            self._save_json(path, **kwargs)
        else:
            raise LeaspyIndividualParamsInputError(
                f"Saving individual parameters to extension '{extension}' is currently not handled. "
                f"Valid extensions are: {self.VALID_IO_EXTENSIONS}."
            )

    @classmethod
    def load(cls, path: str):
        r"""
        Static method that loads the individual parameters (json or csv) existing at the path location

        Parameters
        ----------
        path : :obj:`str`
            Path and file name of the individual parameters.

        Returns
        -------
        :class:`.IndividualParameters`
            Individual parameters object load from the file

        Raises
        ------
        :exc:`.LeaspyIndividualParamsInputError`
            If the provided extension is not `csv` or not `json`.

        Examples
        --------
        >>> ip = IndividualParameters.load('/path/to/individual_parameters_1.json')
        >>> ip2 = IndividualParameters.load('/path/to/individual_parameters_2.csv')
        """
        extension = cls._check_and_get_extension(path)
        if extension not in cls.VALID_IO_EXTENSIONS:
            raise LeaspyIndividualParamsInputError(
                f"Loading individual parameters from extension '{extension}' is currently not handled. "
                f"Valid extensions are: {cls.VALID_IO_EXTENSIONS}."
            )

        if extension == "csv":
            ip = cls._load_csv(path)
        else:
            ip = cls._load_json(path)

        return ip

    @staticmethod
    def _check_and_get_extension(path: str):
        """
        Extract the file extension from a file path.

        Parameters
        ----------
        path : :obj:`str`
            The file path from which to extract the extension

        Returns
        -------
        :obj:`str` or None
            The file extension (e.g. ``'txt'`, ``'csv'``) if present,
            otherwise ``None``.
        """
        _, ext = os.path.splitext(path)
        if len(ext) == 0:
            return None
        else:
            return ext[1:]

    def _save_csv(self, path: str, **kwargs):
        """
        Save the individual parameters to a csv file

        Parameters
        ----------
        path : :obj:`str`
            path where the CSV file will be saved
        **kwargs
            Additional keyword arguments passed to ``pandas.DataFrale.to_csv``
        """
        df = self.to_dataframe()
        df.to_csv(path, **kwargs)

    def _save_json(self, path: str, **kwargs):
        """
        Save individual parameters and related metadata to a JSON file.

        Parameters
        ----------
        path : :obj:`str`
            File path where the JSON data will be saved
        **kwargs 
            Additional keyword arguments passed to ``json.dump``
        """
        json_data = {
            "indices": self._indices,
            "individual_parameters": self._individual_parameters,
            "parameters_shape": self._parameters_shape,
        }

        # Default json.dump kwargs:
        kwargs = {"indent": 2, **kwargs}

        with open(path, "w") as f:
            json.dump(json_data, f, **kwargs)

    @classmethod
    def _load_csv(cls, path: str):
        """
        Load individual parameters from a CSV file

        Parameters
        ----------
        path : :obj:`str`
            Path to the CSV file to load.
        
        Returns
        -------
        :class:`.IndividualParameters`
            Individual parameters object load from the file
        """
        df = pd.read_csv(path, dtype={"ID": IDType}).set_index("ID")
        ip = cls.from_dataframe(df)

        return ip

    @classmethod
    def _load_json(cls, path: str):
        """
        Loads individual parameters and related metadata from a JSON file

        Parameters
        ----------
        path : :obj:`str`
            Path to the JSON file to load.
        
        Returns
        -------
        :class:`.IndividualParameters`
            Individual parameters object load from the file
        """
        with open(path, "r") as f:
            json_data = json.load(f)

        ip = cls()
        ip._indices = json_data["indices"]
        ip._individual_parameters = json_data["individual_parameters"]
        ip._parameters_shape = json_data["parameters_shape"]

        # convert json lists to tuple for shapes
        ip._parameters_shape = {p: tuple(s) for p, s in ip._parameters_shape.items()}

        return ip



"""This module defines the `TensorMCMCSAEM` class."""

from random import shuffle

from leaspy.exceptions import LeaspyAlgoInputError
from leaspy.io.data import Dataset
from leaspy.models import McmcSaemCompatibleModel
from leaspy.variables.specs import (
    IndividualLatentVariable,
    LatentVariableInitType,
    PopulationLatentVariable,
)
from leaspy.variables.state import State

from ..algo_with_annealing import AlgorithmWithAnnealingMixin
from ..algo_with_device import AlgorithmWithDeviceMixin
from ..algo_with_samplers import AlgorithmWithSamplersMixin
from ..base import AlgorithmName
from ..settings import AlgorithmSettings
from .base import FitAlgorithm

__all__ = ["TensorMcmcSaemAlgorithm"]


class TensorMcmcSaemAlgorithm(
    AlgorithmWithDeviceMixin,
    AlgorithmWithAnnealingMixin,
    AlgorithmWithSamplersMixin,
    FitAlgorithm[McmcSaemCompatibleModel, State],
):
    """Main algorithm for MCMC-SAEM.

    Parameters
    ----------
    settings : :class:`~leaspy.algo.AlgorithmSettings`
        MCMC fit algorithm settings

    Attributes
    ----------
    samplers : :obj:`dict` [:obj:`str`, :class:`~leaspy.samplers.AbstractSampler` ]
        Dictionary of samplers per each variable

    random_order_variables : :obj:`bool` (default True)
        This attribute controls whether we randomize the order of variables at each iteration.
        `Article <https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html>`_
        gives a reason on why we should activate this flag.

    temperature : :obj:`float`

    temperature_inv : :obj:`float`
        Temperature and its inverse are modified during algorithm if annealing is used

    See Also
    --------
    :mod:`leaspy.samplers`
    """

    name: AlgorithmName = AlgorithmName.FIT_MCMC_SAEM

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        if not (0.5 < self.algo_parameters["burn_in_step_power"] <= 1):
            raise LeaspyAlgoInputError(
                "The parameter `burn_in_step_power` should be in ]0.5, 1] in order to "
                "have theoretical guarantees on convergence of MCMC-SAEM algorithm."
            )

    def _run(self, model: McmcSaemCompatibleModel, dataset: Dataset, **kwargs) -> State:
        """Main method to run the algorithm.

        Basically, it initializes the :class:`~leaspy.variables.state.State` object,
        updates it using the :meth:`~leaspy.algo.AbstractFitAlgo.iteration` method then returns it.

        Parameters
        ----------
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            The used model. It must be a subclass of :class:`~leaspy.models.McmcSaemCompatibleModel`.

        dataset : :class:`~leaspy.io.data.Dataset`
            Contains the subjects' observations in torch format to speed up computation.

        Returns
        -------
        :class:`~leaspy.variables.state.State` :
            The fitted state.
        """
        with self._device_manager(model, dataset):
            state = self._initialize_algo(model, dataset)
            if self.algo_parameters["progress_bar"]:
                self._display_progress_bar(
                    -1, self.algo_parameters["n_iter"], suffix="iterations"
                )
            for self.current_iteration in range(1, self.algo_parameters["n_iter"] + 1):
                self._iteration(model, state)

                if self.output_manager is not None:
                    # print/plot first & last iteration!
                    # <!> everything that will be printed/saved is AFTER iteration N
                    # (including temperature when annealing...)
                    self.output_manager.iteration(self, model, dataset)

                if self.algo_parameters["progress_bar"]:
                    self._display_progress_bar(
                        self.current_iteration - 1,
                        self.algo_parameters["n_iter"],
                        suffix="iterations",
                    )
        model.fit_metrics = self._get_fit_metrics()
        model_state = state.clone()
        with model_state.auto_fork(None):
            # <!> At the end of the MCMC, population and individual latent variables
            # may have diverged from final model parameters.
            # Thus, we reset population latent variables to their mode
            model_state.put_population_latent_variables(
                LatentVariableInitType.PRIOR_MODE
            )
        model.state = model_state
        return state

    def _initialize_algo(
        self,
        model: McmcSaemCompatibleModel,
        dataset: Dataset,
    ) -> State:
        # TODO? mutualize with perso mcmc algo?
        state = model.state
        with state.auto_fork(None):
            model.put_data_variables(state, dataset)
        # Initialize individual latent variables (population ones should be initialized before)
        model.put_individual_parameters(state, dataset)
        self._initialize_samplers(state, dataset)
        self._initialize_annealing()
        return state

    def _iteration(
        self,
        model: McmcSaemCompatibleModel,
        state: State,
    ) -> None:
        """
        MCMC-SAEM iteration.

        1. Sample : MC sample successively of the population and individual variables
        2. Maximization step : update model parameters from current population/individual variables values.

        Parameters
        ----------
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
        state : :class:`~leaspy.variables.state.State`
        """
        variables = sorted(
            list(state.dag.sorted_variables_by_type[PopulationLatentVariable])
            + list(state.dag.sorted_variables_by_type[IndividualLatentVariable])
        )
        if self.random_order_variables:
            shuffle(variables)
        for variable in variables:
            self.samplers[variable].sample(state, temperature_inv=self.temperature_inv)
        self._maximization_step(model, state)
        self._update_temperature()

    def _maximization_step(self, model: McmcSaemCompatibleModel, state: State):
        """Maximization step as in the EM algorithm.

        In practice parameters are set to current state (burn-in phase), or as a barycenter with previous state.

        Parameters
        ----------
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`

        state : :class:`~leaspy.variables.state.State`
        """
        # TODO/WIP: not 100% clear to me whether model methods should take a state param, or always use its internal state...
        sufficient_statistics = model.compute_sufficient_statistics(state)
        if (
            self._is_burn_in()
            or self.current_iteration == 1 + self.algo_parameters["n_burn_in_iter"]
        ):
            # the maximization step is memoryless (or first iteration with memory)
            self.sufficient_statistics = sufficient_statistics
        else:
            burn_in_step = (
                self.current_iteration - self.algo_parameters["n_burn_in_iter"]
            )  # min = 2, max = n_iter - n_burn_in_iter
            burn_in_step **= -self.algo_parameters["burn_in_step_power"]

            # this new formulation (instead of v + burn_in_step*(sufficient_statistics[k] - v))
            # enables to keep `inf` deltas
            self.sufficient_statistics = {
                k: v * (1.0 - burn_in_step) + burn_in_step * sufficient_statistics[k]
                for k, v in self.sufficient_statistics.items()
            }
        # TODO: use the same method in both cases (<!> very minor differences that might break
        #  exact reproducibility in tests)
        model.update_parameters(
            state, self.sufficient_statistics, burn_in=self._is_burn_in()
        )

    def log_current_iteration(self, state: State):
        if (
            self.is_current_iteration_in_last_n()
            or self.should_current_iteration_be_saved()
        ):
            state.save(
                self.logs.parameter_convergence_path,
                iteration=self.current_iteration,
            )

    def is_current_iteration_in_last_n(self):
        """Return True if current iteration is within the last n realizations defined in logging settings."""
        return (
            self.current_iteration
            > self.algo_parameters["n_iter"] - self.logs.save_last_n_realizations
        )

    def should_current_iteration_be_saved(self):
        """Return True if current iteration should be saved based on log saving periodicity."""
        return (
            self.logs.save_periodicity
            and self.current_iteration % self.logs.save_periodicity == 0
        )



import contextlib

import torch

from leaspy.io.data import Dataset
from leaspy.models import McmcSaemCompatibleModel

from .settings import AlgorithmSettings

__all__ = ["AlgorithmWithDeviceMixin"]


class AlgorithmWithDeviceMixin:
    """Mixin class containing common attributes & methods for algorithms with a torch device.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        The specifications of the algorithm as a :class:`.AlgorithmSettings` instance.

    Attributes
    ----------
    algorithm_device : :obj:`str`
        Valid torch device
    """

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        self.algorithm_device = settings.device
        self._default_algorithm_device = torch.device("cpu")
        self._default_algorithm_tensor_type = "torch.FloatTensor"

    @contextlib.contextmanager
    def _device_manager(self, model: McmcSaemCompatibleModel, dataset: Dataset):
        """
        Context-manager to handle the "ambient device" (i.e. the device used
        to instantiate tensors and perform computations). The provided model
        and dataset will be moved to the device specified for the execution
        at the beginning of the algorithm and moved back to the original
        ('cpu') device at the end of the algorithm. The default tensor type
        will also be set accordingly.

        Parameters
        ----------
        model : :class:`~.models.abstract_model.McmcSaemCompatibleModel`
            The used model.
        dataset : :class:`.Dataset`
            Contains the subjects' observations in torch format to speed up computation.
        """
        algorithm_tensor_type = self._default_algorithm_tensor_type
        if self.algorithm_device != self._default_algorithm_device.type:
            algorithm_device = torch.device(self.algorithm_device)

            model.move_to_device(algorithm_device)
            dataset.move_to_device(algorithm_device)

            algorithm_tensor_type = "torch.cuda.FloatTensor"

        try:
            yield torch.set_default_tensor_type(algorithm_tensor_type)
        finally:
            if self.algorithm_device != self._default_algorithm_device.type:
                model.move_to_device(self._default_algorithm_device)
                dataset.move_to_device(self._default_algorithm_device)

            torch.set_default_tensor_type(self._default_algorithm_tensor_type)
